from wbcore import viewsets
from wbcore.permissions.permissions import InternalUserPermissionMixin

from wbportfolio.filters import PortfolioRoleFilterSet
from wbportfolio.models import PortfolioRole
from wbportfolio.serializers import (
    PortfolioRoleModelSerializer,
    PortfolioRoleProjectModelSerializer,
)

from .configs import (
    PortfolioRoleDisplayConfig,
    PortfolioRoleInstrumentDisplayConfig,
    PortfolioRoleInstrumentEndpointConfig,
    PortfolioRoleInstrumentTitleConfig,
    PortfolioRoleTitleConfig,
)


class PortfolioRoleModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    filterset_class = PortfolioRoleFilterSet

    serializer_class = PortfolioRoleModelSerializer
    queryset = PortfolioRole.objects.select_related(
        "person",
        "instrument",
    )

    ordering_fields = ("start", "end", "weighting", "instrument__computed_str")
    ordering = ["instrument__computed_str"]
    search_fields = ("person__computed_str", "instrument__computed_str")

    display_config_class = PortfolioRoleDisplayConfig
    title_config_class = PortfolioRoleTitleConfig


class PortfolioRoleInstrumentModelViewSet(PortfolioRoleModelViewSet):
    serializer_class = PortfolioRoleProjectModelSerializer

    display_config_class = PortfolioRoleInstrumentDisplayConfig
    title_config_class = PortfolioRoleInstrumentTitleConfig
    endpoint_config_class = PortfolioRoleInstrumentEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(instrument__id=self.kwargs["instrument_id"])



from django.db import models
from django.utils.translation import gettext_lazy as _
from wbcore.models import WBModel


class Reconciliation(WBModel):
    reconciliation_date = models.DateField()

    creator = models.ForeignKey(
        to="authentication.User",
        on_delete=models.PROTECT,
        related_name="+",
    )
    approved_by = models.ForeignKey(
        to="authentication.User",
        on_delete=models.PROTECT,
        null=True,
        blank=True,
        verbose_name=_("Approved By"),
        related_name="+",
    )
    approved_dt = models.DateTimeField(null=True, blank=True, verbose_name=_("Approved Timestamp"))

    class Meta:
        abstract = True



from wbcore.menus import ItemPermission, MenuItem
from wbcore.permissions.shortcuts import is_internal_user

TRADE_MENUITEM = MenuItem(
    label="Trades",
    endpoint="wbportfolio:trade-list",
    permission=ItemPermission(
        permissions=["wbportfolio.view_trade"], method=lambda request: is_internal_user(request.user)
    ),
)
SUBSCRIPTION_REDEMPTION_MENUITEM = MenuItem(
    label="Subscription/Redemption",
    endpoint="wbportfolio:subscriptionredemption-list",
    permission=ItemPermission(
        permissions=["wbportfolio.view_trade"], method=lambda request: is_internal_user(request.user)
    ),
)



import re
import textwrap
from datetime import date

import networkx as nx
import plotly.graph_objects as go
import pydot
from django.db.models import Q

from wbportfolio.models import Portfolio, PortfolioPortfolioThroughModel

from .utils import networkx_graph_to_plotly


class PortfolioGraph:
    def __init__(self, portfolio: Portfolio, val_date: date, **graph_kwargs):
        self.graph = pydot.Dot("Portfolio Tree", strict=True, **graph_kwargs)
        self.base_portfolio = portfolio
        self.discovered_portfolios = set()
        self.val_date = val_date
        self._extend_portfolio_graph(portfolio)

    @classmethod
    def _convert_to_multilines(cls, name: str) -> str:
        lines = textwrap.wrap(name, 30)
        return "\n ".join(lines)

    def _extend_parent_portfolios_to_graph(self, portfolio):
        for parent_portfolio, weighting in portfolio.get_parent_portfolios(self.val_date):
            if parent_portfolio.assets.filter(date=self.val_date).exists():
                self.graph.add_node(
                    pydot.Node(
                        str(parent_portfolio.id),
                        label=self._convert_to_multilines(str(parent_portfolio)),
                        **self._get_node_kwargs(parent_portfolio),
                    )
                )
                self.graph.del_edge((str(portfolio.id), str(parent_portfolio.id)))
                self.graph.add_edge(
                    pydot.Edge(
                        str(portfolio.id),
                        str(parent_portfolio.id),
                        label=f"Invest in ({weighting:.2%})",
                        style="dashed",
                    )
                )
                # composition_edges.append((str(portfolio.id), str(parent_portfolio.id)))
                self._extend_parent_portfolios_to_graph(parent_portfolio)
                self._extend_portfolio_graph(parent_portfolio)

    def _extend_child_portfolios_to_graph(self, portfolio):
        for child_portfolio in portfolio.get_child_portfolios(self.val_date):
            self.graph.add_node(
                pydot.Node(
                    str(child_portfolio.id),
                    label=self._convert_to_multilines(str(child_portfolio)),
                    **self._get_node_kwargs(child_portfolio),
                )
            )
            # we add this edge only if the opposite relationship is not already added
            graph_edge = pydot.Edge(str(child_portfolio.id), str(portfolio.id), label="implements", style="dashed")
            if (graph_edge.get_source(), graph_edge.get_destination()) not in self.graph.obj_dict["edges"]:
                self.graph.add_edge(graph_edge)
            # composition_edges.append((str(child_portfolio.id), str(portfolio.id)))
            self._extend_child_portfolios_to_graph(child_portfolio)

    def _get_node_kwargs(self, portfolio):
        node_args = {
            "shape": "circle",
            "orientation": "45",
        }
        if portfolio == self.base_portfolio:
            node_args.update({"style": "filled", "fillcolor": "lightgrey"})
        else:
            node_args.update(
                {
                    "style": "solid",
                }
            )
        return node_args

    def _extend_portfolio_graph(self, portfolio):
        self.graph.add_node(
            pydot.Node(
                str(portfolio.id),
                label=self._convert_to_multilines(str(portfolio)),
                **self._get_node_kwargs(portfolio),
            )
        )

        self._extend_child_portfolios_to_graph(portfolio)
        self._extend_parent_portfolios_to_graph(portfolio)

        # composition_edges = []
        # if composition_edges:
        #     with self.graph.subgraph(label=f'composition_{portfolio.id}') as a:
        #         a.edges(composition_edges)
        #         a.attr(color='blue')
        #         a.attr(label='Portfolio Composition')
        self.discovered_portfolios.add(portfolio)

        for rel in PortfolioPortfolioThroughModel.objects.filter(
            Q(portfolio=portfolio) | Q(dependency_portfolio=portfolio)
        ):
            self.graph.add_node(
                pydot.Node(
                    str(rel.portfolio.id),
                    label=self._convert_to_multilines(str(rel.portfolio)),
                    **self._get_node_kwargs(rel.portfolio),
                )
            )
            self.graph.add_node(
                pydot.Node(
                    str(rel.dependency_portfolio.id),
                    label=self._convert_to_multilines(str(rel.dependency_portfolio)),
                    **self._get_node_kwargs(rel.dependency_portfolio),
                )
            )
            label = PortfolioPortfolioThroughModel.Type[rel.type].label
            if rel.dependency_portfolio.is_composition:
                label += " (Composition)"

            if rel.portfolio.is_lookthrough and rel.type == PortfolioPortfolioThroughModel.Type.LOOK_THROUGH:
                self.graph.add_edge(
                    pydot.Edge(
                        str(rel.portfolio.id), str(rel.dependency_portfolio.id), label="Look-Through", style="dotted"
                    )
                )
            else:
                self.graph.add_edge(
                    pydot.Edge(
                        str(rel.portfolio.id),
                        str(rel.dependency_portfolio.id),
                        label=label,
                        style="bold",
                    )
                )
            if rel.dependency_portfolio not in self.discovered_portfolios:
                self._extend_portfolio_graph(rel.dependency_portfolio)
            if rel.portfolio not in self.discovered_portfolios:
                self._extend_portfolio_graph(rel.portfolio)

    def to_string(self) -> str:
        return self.graph.to_string()

    def to_networkx(self) -> nx.Graph:
        return nx.drawing.nx_pydot.from_pydot(self.graph)

    def to_plotly(self, **kwargs) -> go.Figure:
        node_labels = {
            node.get_name(): node.obj_dict["attributes"].get("label", node.get_name())
            for node in self.graph.get_node_list()
        }
        return networkx_graph_to_plotly(self.to_networkx(), labels=node_labels, **kwargs)

    def to_svg(self) -> str:
        svg = self.graph.create_svg().decode("utf-8")
        svg_matches = re.findall(r"<svg\b[^<]*(?:(?!<\/svg>)<[^<]*)*<\/svg>", svg, flags=re.DOTALL)
        if svg_matches:
            return svg_matches[0]
        return svg



from contextlib import suppress
from decimal import Decimal

from django.db import models
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.models import WBModel

from wbportfolio.import_export.handlers.portfolio_cash_flow import (
    DailyPortfolioCashFlowImportHandler,
)
from wbportfolio.models.portfolio_cash_targets import PortfolioCashTarget
from wbportfolio.models.portfolio_swing_pricings import PortfolioSwingPricing


class DailyPortfolioCashFlow(ImportMixin, WBModel):
    import_export_handler_class = DailyPortfolioCashFlowImportHandler
    pending = models.BooleanField(default=False)

    portfolio = models.ForeignKey(
        to="wbportfolio.Portfolio",
        related_name="daily_cashflows",
        on_delete=models.CASCADE,
    )
    value_date = models.DateField()

    cash = models.DecimalField(max_digits=19, decimal_places=4, blank=True)
    cash_flow_forecast = models.DecimalField(max_digits=19, decimal_places=4, default=Decimal(0), blank=True)
    total_assets = models.DecimalField(max_digits=19, decimal_places=4, blank=True)

    min_target_cash_pct = models.DecimalField(max_digits=4, decimal_places=4, default=Decimal(0), blank=True)
    target_cash_pct = models.DecimalField(max_digits=4, decimal_places=4, default=Decimal(0), blank=True)
    max_target_cash_pct = models.DecimalField(max_digits=4, decimal_places=4, default=Decimal(0), blank=True)

    swing_pricing = models.ForeignKey(
        to="wbportfolio.PortfolioSwingPricing",
        related_name="cash_flow",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
    )

    estimated_total_assets = models.DecimalField(max_digits=19, decimal_places=4, blank=True)
    cash_flow_asset_ratio = models.DecimalField(max_digits=5, decimal_places=4, blank=True)
    true_cash = models.DecimalField(max_digits=19, decimal_places=4, blank=True)
    cash_pct = models.DecimalField(max_digits=5, decimal_places=4, blank=True)
    true_cash_pct = models.DecimalField(max_digits=5, decimal_places=4, blank=True)
    target_cash = models.DecimalField(max_digits=19, decimal_places=4, blank=True)
    excess_cash = models.DecimalField(max_digits=19, decimal_places=4, blank=True)
    proposed_rebalancing = models.DecimalField(max_digits=19, decimal_places=4, blank=True, default=Decimal(0))
    rebalancing = models.DecimalField(max_digits=19, decimal_places=4, blank=True, default=Decimal(0))

    comment = models.TextField(default="", blank=True)

    def save(self, *args, **kwargs):
        # convert to decimal in case we get floats
        if isinstance(self.cash_flow_forecast, float):
            self.cash_flow_forecast = Decimal(self.cash_flow_forecast)

        if isinstance(self.total_assets, float):
            self.total_assets = Decimal(self.total_assets)

        if isinstance(self.rebalancing, float):
            self.rebalancing = Decimal(self.rebalancing)

        with suppress(PortfolioCashTarget.DoesNotExist):
            cash_target = self.portfolio.cash_targets.filter(valid_date__lte=self.value_date).latest("valid_date")
            self.min_target_cash_pct = cash_target.min_target
            self.target_cash_pct = cash_target.target
            self.max_target_cash_pct = cash_target.max_target

        with suppress(PortfolioSwingPricing.DoesNotExist):
            self.swing_pricing = self.portfolio.swing_pricings.filter(valid_date__lte=self.value_date).latest(
                "valid_date"
            )

        with suppress(self.DoesNotExist):
            if self.total_assets is None or self.pending:
                prev = self.portfolio.daily_cashflows.filter(value_date__lt=self.value_date).latest("value_date")
                if prev.pending:
                    self.total_assets = prev.estimated_total_assets
                else:
                    self.total_assets = prev.total_assets

        if self.total_assets is None:
            self.total_assets = 0

        with suppress(self.DoesNotExist):
            if self.cash is None or self.pending:
                earlier = self.portfolio.daily_cashflows.filter(value_date__lt=self.value_date).latest("value_date")
                self.cash = earlier.true_cash

        if self.cash is None:
            self.cash = 0

        if self.pending and self.rebalancing:
            self.cash -= self.rebalancing

        if self.pending:
            self.estimated_total_assets = self.total_assets + self.cash_flow_forecast
        elif self.estimated_total_assets is None:
            self.estimated_total_assets = self.total_assets

        self.cash_flow_asset_ratio = (
            self.cash_flow_forecast / self.estimated_total_assets
            if self.estimated_total_assets != Decimal(0)
            else Decimal(0)
        )
        self.true_cash = self.cash + self.cash_flow_forecast
        self.cash_pct = (
            self.cash / self.estimated_total_assets if self.estimated_total_assets != Decimal(0) else Decimal(0)
        )
        self.true_cash_pct = (
            self.true_cash / self.estimated_total_assets if self.estimated_total_assets != Decimal(0) else Decimal(0)
        )
        self.target_cash = self.estimated_total_assets * self.target_cash_pct
        self.excess_cash = self.true_cash - self.target_cash

        if self.true_cash_pct < self.min_target_cash_pct or self.true_cash_pct > self.max_target_cash_pct:
            self.proposed_rebalancing = self.excess_cash
        elif self.pending:
            self.proposed_rebalancing = 0

        super().save(*args, **kwargs)

        if self.portfolio.daily_cashflows.filter(value_date__gt=self.value_date).exists():
            self.portfolio.daily_cashflows.filter(value_date__gt=self.value_date).earliest("value_date").save()

    class Meta:
        verbose_name = "Daily Portfolio CashFlow"
        verbose_name_plural = "Daily Portfolio CashFlow"
        constraints = [
            models.UniqueConstraint(fields=["portfolio", "value_date"], name="unique_portfolio_value_date"),
        ]
        permissions = [("administrate_dailyportfoliocashflow", "Can administrate Daily Portfolio CashFlow")]
        notification_types = [
            create_notification_type(
                "wbportfolio.dailyportfoliocashflow.notify_rebalance",
                "Rebalancing suggested",
                "Sends a notification, when the system suggests to rebalance a portfolio due to being outside of the cash target parameters",
                True,
                True,
                False,
            ),
            create_notification_type(
                "wbportfolio.dailyportfoliocashflow.notify_swingpricing",
                "Swing Pricing Notification",
                "Sends a notification, when the system detects a future swing pricing event",
                False,
                False,
                False,
            ),
        ]

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:portfoliocashflow"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:portfoliocashflow-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{portfolio}}: {{cash}}"



from contextlib import suppress
from datetime import date, datetime

import pandas as pd
from celery import chain
from django.db.models import OuterRef, Q, Subquery, Sum
from django.http import HttpResponse
from django.shortcuts import get_object_or_404
from django.utils.dateparse import parse_date
from django.utils.functional import cached_property
from rest_framework.decorators import action
from rest_framework.permissions import IsAdminUser
from rest_framework.response import Response
from rest_framework.reverse import reverse
from wbcore import viewsets
from wbcore.contrib.currency.models import Currency
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.notifications.dispatch import send_notification_as_task
from wbcore.contrib.pandas import fields as pf
from wbcore.permissions.permissions import InternalUserPermissionMixin
from wbfdm.models import Instrument

from wbportfolio.filters import PortfolioFilterSet, PortfolioTreeGraphChartFilterSet
from wbportfolio.models import (
    AssetPosition,
    OrderProposal,
    Portfolio,
    PortfolioPortfolioThroughModel,
    Rebalancer,
    RebalancingModel,
)
from wbportfolio.models.portfolio import compute_lookthrough_as_task
from wbportfolio.serializers import (
    PortfolioModelSerializer,
    PortfolioPortfolioThroughModelSerializer,
    PortfolioRepresentationSerializer,
)

from ..models.graphs.portfolio import PortfolioGraph
from ..models.utils import adjust_quote_as_task
from ..permissions import IsPortfolioManager
from .configs import (
    PortfolioButtonConfig,
    PortfolioDisplayConfig,
    PortfolioEndpointConfig,
    PortfolioPortfolioThroughModelDisplayConfig,
    PortfolioPortfolioThroughModelEndpointConfig,
    PortfolioPreviewConfig,
    PortfolioTitleConfig,
    PortfolioTreeGraphChartEndpointConfig,
    PortfolioTreeGraphChartTitleConfig,
    TopDownPortfolioCompositionPandasDisplayConfig,
    TopDownPortfolioCompositionPandasEndpointConfig,
    TopDownPortfolioCompositionPandasTitleConfig,
)
from .mixins import UserPortfolioRequestPermissionMixin


class PortfolioRepresentationViewSet(InternalUserPermissionMixin, viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:portfolio"

    ordering_fields = ordering = search_fields = ("name",)
    queryset = Portfolio.objects.all()
    serializer_class = PortfolioRepresentationSerializer
    filterset_class = PortfolioFilterSet


class PortfolioModelViewSet(UserPortfolioRequestPermissionMixin, InternalUserPermissionMixin, viewsets.ModelViewSet):
    filterset_class = PortfolioFilterSet
    serializer_class = PortfolioModelSerializer
    queryset = Portfolio.objects.all()

    search_fields = ("currency__key", "name")
    ordering_fields = (
        "name",
        "currency",
        "hedged_currency",
        "updated_at",
        "initial_position_date",
        "last_position_date",
        "last_asset_under_management_usd",
        "last_positions",
        "automatic_rebalancer",
        "last_order_proposal_date",
        "is_manageable",
        "is_tracked",
        "only_weighting",
        "is_lookthrough",
        "is_composition",
    )
    ordering = ["name"]

    display_config_class = PortfolioDisplayConfig
    button_config_class = PortfolioButtonConfig
    title_config_class = PortfolioTitleConfig
    endpoint_config_class = PortfolioEndpointConfig
    preview_config_class = PortfolioPreviewConfig

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .select_related("currency", "hedged_currency")
            .prefetch_related("depends_on", "instruments")
            .annotate(
                last_asset_under_management_usd=Subquery(
                    AssetPosition.objects.filter(portfolio=OuterRef("pk"), date=OuterRef("last_position_date"))
                    .values("portfolio")
                    .annotate(s=Sum("total_value_fx_usd"))
                    .values("s")[:1]
                ),
                last_positions=Subquery(
                    AssetPosition.objects.filter(portfolio=OuterRef("pk"), date=OuterRef("last_position_date"))
                    .values("portfolio")
                    .annotate(s=Sum("shares"))
                    .values("s")[:1]
                ),
                last_order_proposal_date=Subquery(
                    OrderProposal.objects.filter(portfolio=OuterRef("pk"))
                    .order_by("-trade_date")
                    .values("trade_date")[:1]
                ),
            )
        )

    @action(detail=True, methods=["POST"])
    def rebalance(self, request, pk=None):
        if date_str := request.POST.get("trade_date", None):
            trade_date = datetime.strptime(date_str, "%Y-%m-%d")
            order_proposal, _ = OrderProposal.objects.get_or_create(portfolio_id=pk, trade_date=trade_date)
            order_proposal.reset_orders()
            return Response(
                {"endpoint": reverse("wbportfolio:orderproposal-detail", args=[order_proposal.id], request=request)}
            )
        raise HttpResponse("Bad Request", status=400)

    @action(detail=False, methods=["POST"])
    def createmodelportfolio(self, request, pk=None):
        if self.is_portfolio_manager:
            name = request.POST["name"]
            currency_id = request.POST["currency"]
            currency = Currency.objects.get(id=currency_id)
            index_parameters = {}
            Portfolio.create_model_portfolio(name, currency, index_parameters=index_parameters)
            return Response({"send": True})
        raise HttpResponse("Unauthorized", status=403)

    @action(detail=True, methods=["POST"])
    def attachrebalancer(self, request, pk=None):
        try:
            activation_date = datetime.strptime(request.POST["activation_date"], "%Y-%m-%d")
            rebalancing_model = get_object_or_404(RebalancingModel, pk=request.POST["rebalancing_model"])
            frequency = request.POST["frequency"]
            apply_order_proposal_automatically = (
                request.POST.get("apply_order_proposal_automatically", "false") == "true"
            )

            rebalancer, _ = Rebalancer.objects.update_or_create(
                portfolio=self.get_object(),
                defaults={
                    "rebalancing_model": rebalancing_model,
                    "frequency": frequency,
                    "activation_date": activation_date,
                    "apply_order_proposal_automatically": apply_order_proposal_automatically,
                },
            )
            return Response(
                {"endpoint": reverse("wbportfolio:rebalancer-detail", args=[rebalancer.id], request=request)}
            )
        except KeyError:
            return HttpResponse("Bad arguments", status=400)

    @action(detail=True, methods=["POST"], permission_classes=[IsAdminUser])
    def recomputelookthrough(self, request, pk=None):
        portfolio = self.get_object()
        if portfolio.is_lookthrough:
            with suppress(KeyError):
                start = datetime.strptime(request.POST["start"], "%Y-%m-%d")
                end = datetime.strptime(request.POST["end"], "%Y-%m-%d")
                compute_lookthrough_as_task.delay(portfolio.id, start, end)
            return HttpResponse("Ok", status=200)

        return HttpResponse("Bad arguments", status=400)

    @action(detail=False, methods=["POST"], permission_classes=[IsPortfolioManager])
    def adjustquote(self, request, pk=None):
        old_quote = get_object_or_404(Instrument, pk=request.POST["old_quote"])
        new_quote = get_object_or_404(Instrument, pk=request.POST["new_quote"])
        adjust_after = parse_date(request.data["adjust_after"]) if "adjust_after" in request.data else None
        only_portfolio_ids = request.data["only_portfolios"].split(",") if "only_portfolios" in request.data else []

        chain(
            adjust_quote_as_task.si(
                old_quote.id, new_quote.id, adjust_after=adjust_after, only_portfolio_ids=only_portfolio_ids
            ),
            send_notification_as_task.si(
                "wbportfolio.portfolio.action_done",
                f"Quote adjustment from {old_quote} to {new_quote} is done",
                "The associated positions and orders were successfully adjusted",
                request.user.id,
            ),
        ).apply_async()
        return HttpResponse("Ok", status=200)


class PortfolioPortfolioThroughModelViewSet(InternalUserPermissionMixin, viewsets.ModelViewSet):
    serializer_class = PortfolioPortfolioThroughModelSerializer
    queryset = PortfolioPortfolioThroughModel.objects.all()

    search_fields = ["dependency_portfolio__name"]

    display_config_class = PortfolioPortfolioThroughModelDisplayConfig
    endpoint_config_class = PortfolioPortfolioThroughModelEndpointConfig

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .filter(Q(portfolio=self.kwargs["portfolio_id"]) | Q(dependency_portfolio=self.kwargs["portfolio_id"]))
        )


class PortfolioTreeGraphChartViewSet(UserPortfolioRequestPermissionMixin, viewsets.HTMLViewSet):
    filterset_class = PortfolioTreeGraphChartFilterSet
    IDENTIFIER = "wbportfolio:portfolio-treegraphchart"
    queryset = Portfolio.objects.all()

    title_config_class = PortfolioTreeGraphChartTitleConfig
    endpoint_config_class = PortfolioTreeGraphChartEndpointConfig

    def get_queryset(self):
        return super().get_queryset().filter(id=self.kwargs["portfolio_id"])

    @cached_property
    def val_date(self) -> date:
        return parse_date(self.request.GET["date"])

    def get_html(self, queryset) -> str:
        portfolio_graph = PortfolioGraph(self.portfolio, self.val_date, rankdir="LR", size="20,11")
        return portfolio_graph.to_svg()


class TopDownPortfolioCompositionPandasAPIView(UserPortfolioRequestPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "wbportfolio:topdowncomposition"
    queryset = Portfolio.objects.all()

    display_config_class = TopDownPortfolioCompositionPandasDisplayConfig
    title_config_class = TopDownPortfolioCompositionPandasTitleConfig
    endpoint_config_class = TopDownPortfolioCompositionPandasEndpointConfig

    pandas_fields = pf.PandasFields(
        fields=[
            pf.PKField(key="id", label="ID"),
            pf.IntegerField(key="_group_key", label="Group Key"),
            pf.IntegerField(key="parent_row_id", label="Parent Row"),
            pf.CharField(key="instrument", label="Instrument"),
            pf.FloatField(key="effective_weights", label="Effective Weights", precision=2, percent=True),
            pf.FloatField(key="rebalancing_weights", label="Rebalancing Weights", precision=2, percent=True),
        ]
    )

    @cached_property
    def composition_portfolio(self) -> Portfolio | None:
        return self.portfolio.composition_portfolio or self.portfolio

    @cached_property
    def last_rebalancing_date(self) -> date | None:
        if self.composition_portfolio and self.last_effective_date:
            with suppress(OrderProposal.DoesNotExist):
                return (
                    self.composition_portfolio.order_proposals.filter(trade_date__lte=self.last_effective_date)
                    .latest("trade_date")
                    .trade_date
                )

    @cached_property
    def last_effective_date(self) -> date | None:
        if self.composition_portfolio:
            with suppress(AssetPosition.DoesNotExist):
                return self.composition_portfolio.assets.latest("date").date

    def get_dataframe(self, request, queryset, **kwargs):
        df = pd.DataFrame(columns=["id", "parent_row_id", "instrument", "effective_weights", "rebalancing_weights"])

        def _get_parent_instrument_id(portfolio):
            with suppress(AttributeError):
                return str(portfolio.instruments.first().id)

        if (
            self.has_portfolio_access
            and self.composition_portfolio
            and self.last_rebalancing_date
            and self.last_effective_date
        ):
            df = pd.DataFrame(
                map(
                    lambda x: {
                        "instrument": str(x.underlying_instrument.id),
                        "path": "-".join(
                            map(
                                lambda o: _get_parent_instrument_id(o),
                                getattr(x, "path", [self.composition_portfolio]),
                            )
                        ),
                        "effective_weights": x.weighting,
                    },
                    self.composition_portfolio.get_positions(
                        self.last_effective_date, with_intermediary_position=True
                    ),
                ),
                columns=["instrument", "path", "effective_weights"],
            ).set_index(["path", "instrument"])

            if last_rebalancing_date := self.last_rebalancing_date:
                tree_positions_rebalancing_date_df = pd.DataFrame(
                    map(
                        lambda x: {
                            "instrument": str(x.underlying_instrument.id),
                            "path": "-".join(
                                map(
                                    lambda o: _get_parent_instrument_id(o),
                                    getattr(x, "path", [self.composition_portfolio]),
                                )
                            ),
                            "rebalancing_weights": x.weighting,
                        },
                        self.composition_portfolio.get_positions(
                            last_rebalancing_date, with_intermediary_position=True
                        ),
                    ),
                    columns=["instrument", "path", "rebalancing_weights"],
                ).set_index(["path", "instrument"])

                df = pd.concat([df, tree_positions_rebalancing_date_df], axis=1)
            df = df.reset_index()
            df = pd.concat(
                [
                    pd.DataFrame(
                        [
                            {
                                "instrument": str(self.portfolio.instruments.first().id),
                                "path": "",
                                "effective_weights": 1.0,
                                "rebalancing_weights": 1.0,
                            }
                        ]
                    ),
                    df,
                ],
                ignore_index=True,
            )

            df = df.reset_index(names="id")

            def _get_group_key(x):
                return str(int(x)) if not df.loc[df["parent_row_id"] == x, :].empty else None

            def _get_parent_row_id(path):
                s = path.split("-")
                instrument = s[-1]
                parent_path = "-".join(s[:-1]) if len(s) > 1 else ""
                dff = df.loc[(df["instrument"] == instrument) & (df["path"] == parent_path), "id"]
                if not dff.empty:
                    return dff.iloc[0]
                return None

            df["parent_row_id"] = df["path"].apply(lambda x: _get_parent_row_id(x))
            df["_group_key"] = df["id"].apply(lambda x: _get_group_key(x))
            df = df.drop(columns=["path"])
        return df

    def manipulate_dataframe(self, df):
        df["instrument"] = (
            df["instrument"]
            .astype(int)
            .map(dict(Instrument.objects.filter(id__in=df["instrument"]).values_list("id", "name_repr")))
        )
        return df



from django.contrib.messages import warning
from django.core.exceptions import ValidationError
from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.directory.serializers import PersonRepresentationSerializer
from wbcore.serializers import CharField, DefaultFromView

from wbportfolio.models import OrderProposal, Portfolio, RebalancingModel

from .. import PortfolioRepresentationSerializer, RebalancingModelRepresentationSerializer


class OrderProposalRepresentationSerializer(wb_serializers.RepresentationSerializer):
    _detail = wb_serializers.HyperlinkField(reverse_name="wbportfolio:orderproposal-detail")

    class Meta:
        model = OrderProposal
        fields = ("id", "trade_date", "status", "_detail")


class OrderProposalModelSerializer(wb_serializers.ModelSerializer):
    _portfolio = PortfolioRepresentationSerializer(source="portfolio")
    rebalancing_model = wb_serializers.PrimaryKeyRelatedField(queryset=RebalancingModel.objects.all(), required=False)
    _rebalancing_model = RebalancingModelRepresentationSerializer(source="rebalancing_model")
    target_portfolio = wb_serializers.PrimaryKeyRelatedField(
        queryset=Portfolio.objects.all(), write_only=True, required=False
    )
    _target_portfolio = PortfolioRepresentationSerializer(source="target_portfolio")
    trade_date = wb_serializers.DateField(
        read_only=lambda view: not view.new_mode, default=DefaultFromView("default_trade_date")
    )
    _creator = PersonRepresentationSerializer(source="creator")
    _approver = PersonRepresentationSerializer(source="approver")
    execution_status_repr = wb_serializers.SerializerMethodField(label="Status", field_class=CharField, read_only=True)

    def get_execution_status_repr(self, obj):
        repr = obj.execution_status
        if obj.execution_status_detail:
            repr += f" (Custodian: {obj.execution_status_detail})"
        return repr

    def create(self, validated_data):
        target_portfolio = validated_data.pop("target_portfolio", None)
        rebalancing_model = validated_data.get("rebalancing_model", None)
        if request := self.context.get("request"):
            validated_data["creator"] = request.user.profile
        obj = super().create(validated_data)
        try:
            target_portfolio_dto = None
            if target_portfolio:
                target_portfolio_dto = target_portfolio._build_dto(obj.trade_date)
            elif rebalancing_model:
                target_portfolio_dto = obj.get_default_target_portfolio()
            obj.reset_orders(target_portfolio=target_portfolio_dto)
        except ValidationError as e:
            if request := self.context.get("request"):
                warning(request, str(e), extra_tags="auto_close=0")
        return obj

    @wb_serializers.register_only_instance_resource()
    def additional_resources(self, instance, request, user, **kwargs):
        res = {}
        if instance.status == OrderProposal.Status.CONFIRMED:
            res["replay"] = reverse("wbportfolio:orderproposal-replay", args=[instance.id], request=request)
        if instance.status == OrderProposal.Status.DRAFT:
            res["reset"] = reverse("wbportfolio:orderproposal-reset", args=[instance.id], request=request)
            res["normalize"] = reverse("wbportfolio:orderproposal-normalize", args=[instance.id], request=request)
        if instance.status == OrderProposal.Status.DRAFT or instance.can_be_confirmed:
            res["refresh_return"] = reverse(
                "wbportfolio:orderproposal-refreshreturn", args=[instance.id], request=request
            )
        res["orders"] = reverse(
            "wbportfolio:orderproposal-order-list",
            args=[instance.id],
            request=request,
        )
        return res

    class Meta:
        model = OrderProposal
        only_fsm_transition_on_instance = True
        percent_fields = ["total_cash_weight"]
        fields = (
            "id",
            "trade_date",
            "portfolio",
            "_portfolio",
            "total_cash_weight",
            "comment",
            "status",
            "min_order_value",
            "min_weighting",
            "_rebalancing_model",
            "rebalancing_model",
            "target_portfolio",
            "_target_portfolio",
            "creator",
            "approver",
            "_creator",
            "_approver",
            "execution_status",
            "execution_status_detail",
            "execution_comment",
            "execution_status_repr",
            "_additional_resources",
        )


class ReadOnlyOrderProposalModelSerializer(OrderProposalModelSerializer):
    class Meta(OrderProposalModelSerializer.Meta):
        read_only_fields = OrderProposalModelSerializer.Meta.fields



import enum
from dataclasses import asdict, dataclass, field, fields
from datetime import date
from datetime import date as date_lib
from decimal import Decimal
from typing import Self

import pandas as pd
from django.core.exceptions import ValidationError

from wbportfolio.order_routing import ExecutionInstruction


@dataclass(frozen=True)
class Valuation:
    instrument: int
    net_value: Decimal
    outstanding_shares: Decimal = Decimal(0)


@dataclass()
class Position:
    underlying_instrument: int
    weighting: Decimal
    date: date_lib

    daily_return: Decimal = Decimal("0")
    currency: int | None = None
    instrument_type: int | None = None
    asset_valuation_date: date_lib | None = None
    portfolio_created: int = None
    exchange: int = None
    is_estimated: bool = False
    country: int = None
    shares: Decimal | None = None
    is_cash: bool = False
    primary_classification: int = None
    favorite_classification: int = None
    market_capitalization_usd: float = None
    currency_fx_rate: float = 1
    market_share: float = None
    daily_liquidity: float = None
    volume_usd: float = None
    price: float = None

    def __post_init__(self):
        self.daily_return = round(self.daily_return, 16)
        self.weighting = round(self.weighting, 8)

    def __add__(self, other):
        return Position(
            weighting=self.weighting + other.weighting,
            shares=self.shares + other.shares if (self.shares is not None and other.shares is not None) else None,
            **{f.name: getattr(self, f.name) for f in fields(Position) if f.name not in ["weighting", "shares"]},
        )

    def copy(self, **kwargs):
        attrs = {f.name: getattr(self, f.name) for f in fields(Position)}
        attrs.update(kwargs)
        return Position(**attrs)


@dataclass(frozen=True)
class Order:
    class AssetType(enum.Enum):
        EQUITY = "EQUITY"
        AMERICAN_DEPOSITORY_RECEIPT = "AMERICAN_DEPOSITORY_RECEIPT"

    id: int | str
    trade_date: date
    target_weight: float

    # Instrument identifier
    asset_class: AssetType
    refinitiv_identifier_code: str | None = None
    bloomberg_ticker: str | None = None
    sedol: str | None = None

    weighting: float | None = None
    target_shares: float | None = None
    shares: float | None = None
    execution_price: float | None = None
    execution_instruction: ExecutionInstruction = ExecutionInstruction.MARKET_ON_CLOSE.value
    execution_instruction_parameters: dict | None = None
    comment: str = ""


@dataclass()
class Trade:
    underlying_instrument: int
    instrument_type: int
    currency: int
    date: date_lib
    price: Decimal
    effective_weight: Decimal
    target_weight: Decimal
    currency_fx_rate: Decimal = Decimal("1")
    effective_shares: Decimal = Decimal("0")
    target_shares: Decimal = Decimal("0")
    daily_return: Decimal = Decimal("0")
    effective_quantization_error: Decimal = Decimal("0")
    target_quantization_error: Decimal = Decimal("0")

    id: int | None = None
    is_cash: bool = False

    def __post_init__(self):
        self.effective_weight = round(self.effective_weight, 8)
        # ensure a trade target weight cannot be lower than 0
        self.target_weight = max(round(self.target_weight, 8), Decimal("0"))
        self.daily_return = round(self.daily_return, 16)

    def __add__(self, other):
        return Trade(
            underlying_instrument=self.underlying_instrument,
            effective_weight=self.effective_weight,
            target_weight=self.target_weight + other.target_weight,
            effective_shares=self.effective_shares,
            target_shares=self.target_shares + other.target_shares,
            daily_return=self.daily_return,
            **{
                f.name: getattr(self, f.name)
                for f in fields(Trade)
                if f.name
                not in [
                    "effective_weight",
                    "target_weight",
                    "effective_shares",
                    "target_shares",
                    "underlying_instrument",
                    "daily_return",
                ]
            },
        )

    def copy(self, **kwargs):
        attrs = {f.name: getattr(self, f.name) for f in fields(Trade)}
        attrs.update(kwargs)
        return Trade(**attrs)

    @property
    def delta_weight(self) -> Decimal:
        return (self.target_weight + self.target_quantization_error) - self.effective_weight

    @property
    def delta_shares(self) -> Decimal:
        return self.target_shares - self.effective_shares

    @property
    def price_fx_portfolio(self) -> Decimal:
        return self.price * self.currency_fx_rate

    def validate(self):
        return True
        # if self.effective_weight < 0 or self.effective_weight > 1.0:
        #     raise ValidationError("Effective Weight needs to be in range [0, 1]")
        # if self.target_weight < 0 or self.target_weight > 1.0:
        #     raise ValidationError("Target Weight needs to be in range [0, 1]")

    def normalize_target(
        self, factor: Decimal | None = None, target_shares: Decimal | int | None = None, target_weight: Decimal = None
    ):
        if factor is None:
            if target_shares is not None:
                factor = target_shares / self.target_shares if self.target_shares else Decimal("1")
            elif target_weight is not None:
                factor = target_weight / self.target_weight if self.target_weight else Decimal("1")
            else:
                raise ValueError("Target weight and shares cannot be both None")
        return self.copy(target_weight=self.target_weight * factor, target_shares=self.target_shares * factor)


@dataclass(frozen=True)
class TradeBatch:
    trades: list[Trade]
    trades_map: dict[Trade] = field(init=False, repr=False)

    def __post_init__(self):
        trade_map = {}
        if self.total_effective_weight and (quant_error := Decimal("1") - self.total_effective_weight):
            self.largest_effective_order.effective_quantization_error = quant_error
        if self.total_target_weight and (quant_error := Decimal("1") - self.total_target_weight):
            self.largest_effective_order.target_quantization_error = quant_error
        for trade in self.trades:
            if trade.underlying_instrument in trade_map:
                trade_map[trade.underlying_instrument] += trade
            else:
                trade_map[trade.underlying_instrument] = trade
        object.__setattr__(self, "trades_map", trade_map)

    @property
    def largest_effective_order(self) -> Trade:
        return max(self.trades, key=lambda obj: obj.effective_weight)

    @property
    def total_target_weight(self) -> Decimal:
        return sum([trade.target_weight for trade in self.trades], Decimal("0"))

    @property
    def total_effective_weight(self) -> Decimal:
        return sum([trade.effective_weight for trade in self.trades], Decimal("0"))

    @property
    def total_abs_delta_weight(self) -> Decimal:
        return sum([abs(trade.delta_weight) for trade in self.trades], Decimal("0"))

    def __add__(self, other):
        return TradeBatch(tuple(self.trades + other.trades))

    def __len__(self):
        return len(self.trades)

    def __iter__(self):
        return iter(self.trades_map.values())

    def validate(self):
        if round(float(self.total_target_weight), 4) != 1:  # we do that to remove decimal over precision
            raise ValidationError(f"Total Weight cannot be different than 1 ({float(self.total_target_weight)})")


@dataclass(frozen=True)
class Portfolio:
    positions: list[Position] | list
    positions_map: dict[int, Position] = field(init=False, repr=False)

    def __post_init__(self):
        positions_map = {}

        for pos in self.positions:
            if pos.underlying_instrument in positions_map:
                positions_map[pos.underlying_instrument] += pos
            else:
                positions_map[pos.underlying_instrument] = pos
        object.__setattr__(self, "positions_map", positions_map)

    @property
    def total_weight(self):
        return sum([pos.weighting for pos in self.positions])

    @property
    def total_shares(self):
        return sum([pos.target_shares for pos in self.positions if pos.target_shares is not None])

    def to_df(self, exclude_cash: bool = False) -> pd.DataFrame:
        return pd.DataFrame([asdict(pos) for pos in self.positions if not exclude_cash or not pos.is_cash])

    def to_dict(self) -> dict[int, Decimal]:
        return {underlying_instrument: pos.weighting for underlying_instrument, pos in self.positions_map.items()}

    def get_orders(self, target_portfolio: Self) -> TradeBatch:
        instruments = self.positions_map.copy()
        instruments.update(target_portfolio.positions_map)

        trades: list[Trade] = []
        for instrument_id, pos in instruments.items():
            effective_weight = target_weight = 0
            effective_shares = target_shares = 0
            daily_return = 0
            price = Decimal("0")
            is_cash = False
            trade_date = None
            if effective_pos := self.positions_map.get(instrument_id, None):
                effective_weight = effective_pos.weighting
                effective_shares = effective_pos.shares
                daily_return = effective_pos.daily_return
                is_cash = effective_pos.is_cash
                price = effective_pos.price
                trade_date = effective_pos.date

            if target_pos := target_portfolio.positions_map.get(instrument_id, None):
                target_weight = target_pos.weighting
                is_cash = target_pos.is_cash
                if target_pos.shares is not None:
                    target_shares = target_pos.shares
                if target_pos.price:
                    price = target_pos.price
                trade_date = target_pos.date

            trade = Trade(
                underlying_instrument=instrument_id,
                effective_weight=effective_weight,
                target_weight=target_weight,
                effective_shares=effective_shares,
                target_shares=target_shares,
                date=trade_date,
                instrument_type=pos.instrument_type,
                currency=pos.currency,
                price=Decimal(price) if price else Decimal("0"),
                currency_fx_rate=Decimal(pos.currency_fx_rate),
                daily_return=Decimal(daily_return),
                is_cash=is_cash,
            )
            trades.append(trade)
        return TradeBatch(trades)

    def __len__(self):
        return len(self.positions)

    def __bool__(self):
        return len(self.positions) > 0

    def normalize_cash(self, target_cash_weight: Decimal):
        """
        Normalize the instantiate portfolio so that the sum of the cash position equals to the new target cash position
        """
        positions = list(filter(lambda pos: not pos.is_cash, self.positions))
        cash_position = next(filter(lambda pos: pos.is_cash, self.positions), None)
        total_non_cash_weight = sum(map(lambda pos: pos.weighting, positions))
        target_weight = Decimal("1") - target_cash_weight
        for pos in positions:
            pos.weighting = pos.weighting * target_weight / total_non_cash_weight
        if cash_position:
            cash_position.weighting = target_cash_weight
            positions.append(cash_position)
        return Portfolio(positions)



import json

import numpy as np
import pandas as pd
from django.utils.dateparse import parse_date

BASE_MAPPING = {
    "instrument": "underlying_instrument__name",
    "ric": "underlying_instrument__refinitiv_identifier_code",
    "bbTicker": "underlying_instrument__ticker",
    "isin": "underlying_instrument__isin",
    "assetClass": "underlying_instrument__instrument_type",
    "direction": "transaction_subtype",
    "currency": "currency__key",
    "quantityTraded": "shares",
    "localPrice": "price",
    "fxMultiplier": "currency_fx_rate",
    "tradeDate": "book_date",
}


def parse(import_source):
    content = json.load(import_source.file)
    data = []
    if (rebalances := content.get("rebalances", None)) and (isin := content.get("isin", None)):
        for rebalance_data in rebalances:
            rebalancing_date = parse_date(rebalance_data["rebalanceDate"])
            df = pd.DataFrame(rebalance_data["items"]).replace([np.inf, -np.inf, np.nan], None)
            df = df.rename(columns=BASE_MAPPING)
            df = df.drop(columns=df.columns.difference(BASE_MAPPING.values()))
            df["underlying_instrument__instrument_type"] = df["underlying_instrument__instrument_type"].str.lower()
            df["transaction_date"] = rebalancing_date.strftime("%Y-%m-%d")
            df["portfolio__isin"] = isin
            df.loc[df["book_date"].isnull(), "book_date"] = df.loc[df["book_date"].isnull(), "transaction_date"]
            df["currency_fx_rate"] = 1 / df["currency_fx_rate"]
            df.loc[df["price"].isnull() & (df["underlying_instrument__instrument_type"] == "cash"), "price"] = 1.0
            df["underlying_instrument__currency__key"] = df["currency__key"]
            data.extend(df.to_dict(orient="records"))
    return {"data": data}



# Generated by Django 5.0.13 on 2025-03-20 08:19

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0074_alter_rebalancer_frequency_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='portfolio',
            name='initial_position_date',
            field=models.DateField(blank=True, null=True, verbose_name='Last Position Date'),
        ),
        migrations.AddField(
            model_name='portfolio',
            name='last_position_date',
            field=models.DateField(blank=True, null=True, verbose_name='Last Position Date'),
        ),
        migrations.AlterField(
            model_name='trade',
            name='transaction_subtype',
            field=models.CharField(choices=[('REBALANCE', 'Rebalance'), ('DECREASE', 'Decrease'), ('INCREASE', 'Increase'), ('SUBSCRIPTION', 'Subscription'), ('REDEMPTION', 'Redemption'), ('BUY', 'Buy'), ('SELL', 'Sell'), ('NO_CHANGE', 'No Change')], default='BUY', max_length=32, verbose_name='Trade Type'),
        ),
    ]



# Generated by Django 5.0.14 on 2025-08-04 09:30

from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0085_order_desired_target_weight'),
    ]

    operations = [
        migrations.AddField(
            model_name='orderproposal',
            name='total_cash_weight',
            field=models.DecimalField(decimal_places=4, default=Decimal('0'), help_text='The desired percentage for the cash component. The remaining percentage (100% minus this value) will be allocated to total target weighting. Default is 0%.', max_digits=5, verbose_name='Total Cash Weight'),
        ),
    ]



from contextlib import suppress

from rest_framework.reverse import reverse
from wbcore import serializers as wb_serializers
from wbcore.contrib.icons import WBIcon
from wbcore.enums import RequestType
from wbcore.metadata.configs import buttons as bt
from wbcore.metadata.configs.buttons.view_config import ButtonViewConfig
from wbcore.metadata.configs.display import create_simple_display

from wbportfolio.models import OrderProposal, Portfolio
from wbportfolio.order_routing import ExecutionStatus
from wbportfolio.serializers import PortfolioRepresentationSerializer


class NormalizeSerializer(wb_serializers.Serializer):
    total_cash_weight = wb_serializers.FloatField(default=0, precision=4, percent=True)


class ResetSerializer(wb_serializers.Serializer):
    use_desired_target_weight = wb_serializers.BooleanField(
        default=False,
        label="Use initial target weight",
        help_text="If True, the target weight used will be the value at the time the order proposal was submitted (as it may have changed due to previous modifications). If False, the delta weight will be set to 0 instead.",
    )


class ExecuteSerializer(wb_serializers.Serializer):
    prioritize_target_weight = wb_serializers.BooleanField(
        default=False,
        label="Prioritize Target Weight",
        help_text="If True, we will communicate to the custodian to prioritize target weight in case both shares and target weight are communicated.",
    )


class OrderProposalButtonConfig(ButtonViewConfig):
    def get_custom_list_instance_buttons(self):
        buttons = []
        with suppress(AttributeError, AssertionError):
            order_proposal = self.view.get_object()
            if order_proposal.can_execute(self.request.user):
                buttons.append(
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:order",),
                        icon=WBIcon.DEAL_MONEY.icon,
                        endpoint=reverse("wbportfolio:orderproposal-execute", args=[order_proposal.id]),
                        label="Execute",
                        action_label="Execute",
                        description_fields="<p>Execute the orders through the setup custodian.</p>",
                        serializer=ExecuteSerializer,
                        instance_display=create_simple_display([["prioritize_target_weight"]]),
                    )
                )

            elif order_proposal.status == OrderProposal.Status.EXECUTION and order_proposal.execution_status in [
                ExecutionStatus.PENDING,
                ExecutionStatus.IN_DRAFT,
            ]:
                buttons.append(
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:order",),
                        icon=WBIcon.PREVIOUS.icon,
                        endpoint=reverse("wbportfolio:orderproposal-cancelexecution", args=[order_proposal.id]),
                        label="Cancel Execution",
                        action_label="Cancel Execution",
                        description_fields="<p>Cancel the current requested execution. Time sensitive operation.</p>",
                    )
                )
                buttons.append(
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:order",),
                        icon=WBIcon.REFRESH.icon,
                        endpoint=reverse("wbportfolio:orderproposal-updateexecutionstatus", args=[order_proposal.id]),
                        label="Update Execution Status",
                        action_label="Update Execution Status",
                        description_fields="<p>Update Execution Status.<p>",
                    )
                )
            if order_proposal.can_be_confirmed and order_proposal.portfolio.is_model:

                class PushModelChangeSerializer(wb_serializers.Serializer):
                    only_for_portfolio_ids = wb_serializers.PrimaryKeyRelatedField(
                        label="Only for Portfolios", queryset=Portfolio.objects.all()
                    )
                    _only_for_portfolio_ids = PortfolioRepresentationSerializer(
                        source="only_for_portfolio_ids",
                        many=True,
                        filter_params={"modeled_after": order_proposal.portfolio.id},
                    )
                    approve_automatically = wb_serializers.BooleanField(
                        default=False,
                        help_text="True if you want all created orders to be automatically move to the approve state.",
                    )

                buttons.append(
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:order",),
                        endpoint=reverse("wbportfolio:orderproposal-pushmodelchange", args=[order_proposal.id]),
                        icon=WBIcon.BROADCAST.icon,
                        label="Push Model Changes",
                        description_fields=f"""
                        Push this rebalancing to all portfolios that are modeled after {order_proposal.portfolio}
                        """,
                        action_label="Push Model Changes",
                        title="Push Model Changes",
                        serializer=PushModelChangeSerializer,
                        instance_display=create_simple_display(
                            [["only_for_portfolio_ids"], ["approve_automatically"]]
                        ),
                    )
                )
            if order_proposal.status == OrderProposal.Status.DRAFT:
                buttons.append(
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:order",),
                        endpoint=reverse("wbportfolio:orderproposal-refreshpretradechecks", args=[order_proposal.id]),
                        icon=WBIcon.RUNNING.icon,
                        label="Evaluate Pre-Trade Checks",
                        action_label="Checks are running.",
                        title="Evaluate Pre-Trade Checks",
                    )
                )
        buttons.append(
            bt.DropDownButton(
                label="Tools",
                buttons=(
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:order",),
                        key="replay",
                        icon=WBIcon.SYNCHRONIZE.icon,
                        label="Replay Orders",
                        description_fields="""
                            <p>Replay Orders. It will recompute all assets positions until next order proposal day (or today otherwise) </p>
                            """,
                        action_label="Replay Order",
                        title="Replay Order",
                    ),
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:order",),
                        key="reset",
                        icon=WBIcon.REGENERATE.icon,
                        label="Reset Orders",
                        description_fields="""
                                <p><strong>Warning:</strong>This action will reset the order delta weight to either 0 or the difference between the previous weight and the locked target weight, depending on the user’s choice.</p>
                                <p><strong>Note:</strong>This operation will change the current delta weights and cannot be undone</p>
                                """,
                        action_label="Reset Orders",
                        title="Reset Orders",
                        serializer=ResetSerializer,
                        instance_display=create_simple_display([["use_desired_target_weight"]]),
                    ),
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:order",),
                        key="normalize",
                        icon=WBIcon.EDIT.icon,
                        label="Normalize Orders",
                        description_fields="""
                                <p>Make sure all orders normalize to a total target weight of (1 - {{total_cash_weight}})</p>
                                """,
                        action_label="Normalize Orders",
                        title="Normalize Orders",
                        serializer=NormalizeSerializer,
                        instance_display=create_simple_display([["total_cash_weight"]]),
                    ),
                    bt.ActionButton(
                        method=RequestType.PATCH,
                        identifiers=("wbportfolio:order",),
                        key="refresh_return",
                        icon=WBIcon.REFRESH.icon,
                        label="Refresh Returns & Price",
                        action_label="Refresh Returns & Price",
                        title="Refresh Returns & Price",
                    ),
                ),
            )
        )
        return set(buttons)

    def get_custom_instance_buttons(self):
        return self.get_custom_list_instance_buttons()



from wbcore.metadata.configs import display as dp
from wbcore.metadata.configs.display.instance_display.utils import repeat_field
from wbcore.metadata.configs.display.view_config import DisplayViewConfig


class AccountReconciliationDisplayViewConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        return dp.ListDisplay(
            fields=[
                dp.Field(key="reconciliation_date", label="Reconciliation Date"),
                dp.Field(key="account", label="Account"),
                dp.Field(key="creator", label="Creator"),
                dp.Field(key="approved_by", label="Approved By"),
                dp.Field(key="approved_dt", label="Approved Timestamp"),
            ],
            formatting=[
                dp.Formatting(
                    column="approved_dt",
                    formatting_rules=[dp.FormattingRule(condition=("∃", True), style={"backgroundColor": "#77DD77"})],
                ),
            ],
            legends=[dp.Legend(items=[dp.LegendItem(label="Agreed Reconciliations", icon="#77DD77")])],
        )

    def get_instance_display(self) -> dp.Display:
        return dp.Display(
            pages=[
                dp.Page(
                    layouts={
                        dp.default(): dp.Layout(
                            grid_template_areas=[
                                ["reconciliation_date", "account", "creator", "approved_by", "approved_dt"],
                                [repeat_field(5, "lines")],
                            ],
                            grid_template_rows=["min-content", "1fr"],
                            inlines=[dp.Inline(key="lines", endpoint="lines")],
                        )
                    }
                )
            ]
        )


class AccountReconciliationLineDisplayViewConfig(DisplayViewConfig):
    def get_list_display(self) -> dp.ListDisplay:
        difference = [
            dp.FormattingRule(condition=("<", 0), style={"fontWeight": "bold", "color": "#FF6961"}),
            dp.FormattingRule(condition=(">", 0), style={"fontWeight": "bold", "color": "green"}),
        ]
        editable = [dp.FormattingRule(style={"fontWeight": "bold"})]
        equal = [dp.FormattingRule(condition=("==", False, "is_equal"), style={"backgroundColor": "orange"})]
        border_left = [
            dp.FormattingRule(
                style={
                    "borderLeft": "1px solid #bdc3c7",
                }
            )
        ]
        border_right = [
            dp.FormattingRule(
                style={
                    "borderRight": "1px solid #bdc3c7",
                }
            )
        ]

        return dp.ListDisplay(
            hide_control_bar=True,
            fields=[
                dp.Field(
                    key="product",
                    label="Product",
                    width=450,
                ),
                dp.Field(
                    key="price_date",
                    label="Date",
                    width=100,
                ),
                dp.Field(
                    label="Our Calculations",
                    key="system",
                    children=[
                        dp.Field(
                            key="shares",
                            label="Shares",
                            width=90,
                        ),
                        dp.Field(
                            key="nominal_value",
                            label="Nominal",
                            width=90,
                        ),
                        dp.Field(
                            key="price",
                            label="Price",
                            width=100,
                        ),
                        dp.Field(
                            key="assets_under_management",
                            label="AuM",
                            width=120,
                        ),
                    ],
                ),
                dp.Field(
                    label="Your Input (Please adjust the columns in bold if necessary)",
                    key="confirmation",
                    children=[
                        dp.Field(
                            key="shares_external",
                            label="Shares",
                            width=90,
                            formatting_rules=[*equal, *editable, *border_left],
                        ),
                        dp.Field(
                            key="nominal_value_external",
                            label="Nominal",
                            width=90,
                            formatting_rules=[*equal, *editable],
                        ),
                        dp.Field(
                            key="price",
                            label="Price",
                            width=100,
                        ),
                        dp.Field(
                            key="assets_under_management_external",
                            label="AuM",
                            width=120,
                            formatting_rules=[*equal, *border_right],
                        ),
                    ],
                ),
                dp.Field(
                    label="Differences",
                    key="differences",
                    children=[
                        dp.Field(key="pct_diff", label="Shares", width=90, formatting_rules=difference),
                        dp.Field(key="shares_diff", label="Shares", width=90, formatting_rules=difference),
                        dp.Field(key="nominal_value_diff", label="Nominal", width=90, formatting_rules=difference),
                        dp.Field(
                            key="assets_under_management_diff",
                            label="AuM",
                            width=120,
                            formatting_rules=difference,
                        ),
                    ],
                ),
            ],
        )



import json
from contextlib import suppress
from datetime import date, timedelta
from decimal import Decimal

import numpy as np
import pandas as pd
import plotly.graph_objects as go
from django.apps import apps
from django.contrib.messages import warning
from django.db.models import (
    BooleanField,
    Case,
    CharField,
    DecimalField,
    ExpressionWrapper,
    F,
    FloatField,
    OuterRef,
    Q,
    Subquery,
    Sum,
    Value,
    When,
)
from django.db.models.functions import Coalesce, Concat, Least
from django.shortcuts import get_object_or_404
from django.utils.dateparse import parse_date
from django.utils.functional import cached_property
from rest_framework import filters, status, viewsets
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets as wb_viewsets
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.enums import WidgetType
from wbcore.filters import DjangoFilterBackend
from wbcore.serializers import decorator
from wbcore.utils.date import get_date_interval_from_request
from wbcore.utils.strings import format_number
from wbcrm.models.accounts import Account
from wbfdm.models import ClassificationGroup, InstrumentPrice
from wbfdm.preferences import get_default_classification_group

from wbportfolio.analysis.claims import ConsolidatedTradeSummary
from wbportfolio.filters import (
    ClaimFilter,
    ConsolidatedTradeSummaryTableFilterSet,
    CumulativeNNMChartFilter,
    CustomerAPIFilter,
    CustomerClaimFilter,
    CustomerClaimGroupByFilter,
    DistributionNNMChartFilter,
    NegativeTermimalAccountPerProductFilterSet,
    ProfitAndLossPandasFilter,
)
from wbportfolio.models import Product, Trade
from wbportfolio.models.transactions.claim import Claim, ClaimGroupbyChoice
from wbportfolio.preferences import get_monthly_nnm_target
from wbportfolio.serializers import (
    ClaimAPIModelSerializer,
    ClaimCustomerModelSerializer,
    ClaimModelSerializer,
    ClaimRepresentationSerializer,
    ClaimTradeModelSerializer,
    NegativeTermimalAccountPerProductModelSerializer,
)

from ..configs.buttons import (
    ClaimTradeButtonConfig,
    ConsolidatedTradeSummaryButtonConfig,
)
from ..configs.buttons.claims import TransferTradeSerializer
from ..configs.display import (
    ClaimDisplayConfig,
    ConsolidatedTradeSummaryDisplayConfig,
    NegativeTermimalAccountPerProductDisplayConfig,
    ProfitAndLossPandasDisplayConfig,
)
from ..configs.endpoints import (
    ClaimAccountEndpointConfig,
    ClaimEndpointConfig,
    ClaimEntryEndpointConfig,
    ClaimProductEndpointConfig,
    ClaimTradeEndpointConfig,
    ConsolidatedTradeSummaryDistributionChartEndpointConfig,
    ConsolidatedTradeSummaryEndpointConfig,
    CumulativeNNMChartEndpointConfig,
    NegativeTermimalAccountPerProductEndpointConfig,
    ProfitAndLossPandasEndpointConfig,
)
from ..configs.titles import (
    ClaimAccountTitleConfig,
    ClaimEntryTitleConfig,
    ClaimProductTitleConfig,
    ClaimTitleConfig,
    ClaimTradeTitleConfig,
    ConsolidatedTradeSummaryDistributionChartTitleConfig,
    ConsolidatedTradeSummaryTitleConfig,
    CumulativeNNMChartTitleConfig,
    NegativeTermimalAccountPerProductTitleConfig,
    ProfitAndLossPandasTitleConfig,
)
from .mixins import ClaimPermissionMixin


class ClaimAPIModelViewSet(ClaimPermissionMixin, viewsets.ModelViewSet):
    filter_backends = [DjangoFilterBackend]
    filterset_class = CustomerAPIFilter
    queryset = Claim.objects.select_related("product", "claimant")
    serializer_class = ClaimAPIModelSerializer


class ClaimRepresentationViewSet(ClaimPermissionMixin, wb_viewsets.RepresentationViewSet):
    IDENTIFIER = "wbportfolio:claim"

    filter_backends = (filters.OrderingFilter, filters.SearchFilter)
    serializer_class = ClaimRepresentationSerializer
    queryset = Claim.objects.all()

    ordering_fields = ("title",)
    search_fields = (
        "product__name",
        "product__isin",
        "product__ticker",
        "claimant__computed_str",
    )
    ordering = ["id"]

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .exclude(status=Claim.Status.WITHDRAWN)
            .select_related("product", "claimant", "account")
        )


class ClaimModelViewSet(ClaimPermissionMixin, wb_viewsets.ModelViewSet):
    IDENTIFIER = "wbportfolio:claim"

    serializer_class = ClaimModelSerializer
    queryset = Claim.objects.exclude(status=Claim.Status.WITHDRAWN)

    search_fields = [
        "claimant__computed_str",
        "product__name",
        "product__isin",
        "bank",
        "account__title",
        "trade_comment",
    ]
    ordering_fields = (
        "date",
        "product__name",
        "claimant",
        "account__title",
        "trade_comment",
        "shares",
        "last_nav",
        "total_value",
        "total_value_usd",
        "bank",
        "creator",
        "reference",
    )
    ordering = ["id", "-date"]

    display_config_class = ClaimDisplayConfig
    title_config_class = ClaimTitleConfig
    endpoint_config_class = ClaimEndpointConfig

    def get_filterset_class(self, request):
        profile = request.user.profile
        if profile.is_internal or request.user.is_superuser:
            return ClaimFilter
        return CustomerClaimFilter

    def get_messages(self, request, queryset=None, paginated_queryset=None, instance=None, initial=False):
        if instance and instance.product and instance.account:
            sum_shares = Claim.objects.exclude(status=Claim.Status.WITHDRAWN).filter(
                product=instance.product, account=instance.account
            ).aggregate(s=Sum("shares"))["s"] or Decimal(0.0)
            if sum_shares < 0:
                warning(
                    request,
                    f"The total shares balance for this account and for this product is negative of {sum_shares}",
                )

    def get_aggregates(self, queryset, paginated_queryset):
        return {
            "shares": {"Σ": format_number(queryset.aggregate(s=Sum("shares"))["s"] or 0, decimal=4)},
        }

    def get_serializer_class(self):
        profile = self.request.user.profile
        if profile.is_internal or self.request.user.is_superuser:
            return super().get_serializer_class()
        return ClaimCustomerModelSerializer

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .annotate(
                currency=F("product__currency__symbol"),
                last_nav=InstrumentPrice.subquery_closest_value(
                    "net_value", instrument_pk_name="product", date_lookup="exact"
                ),
                fx_rate=CurrencyFXRates.get_fx_rates_subquery("date", lookup_expr="exact"),
                total_value=F("last_nav") * F("shares"),
                total_value_usd=ExpressionWrapper(F("fx_rate") * F("total_value"), output_field=FloatField()),
                trade_claimed_shares=Coalesce(
                    Subquery(
                        Claim.objects.filter(Q(status=Claim.Status.APPROVED) & Q(trade=OuterRef("trade")))
                        .values("trade")
                        .annotate(s=Sum("shares"))
                        .values("s")[:1]
                    ),
                    Decimal(0),
                ),
                quantity=Case(
                    When(as_shares=True, then=F("shares")),
                    When(as_shares=False, then=F("nominal_amount")),
                    default=None,
                    output_field=DecimalField(max_digits=15, decimal_places=2),
                ),
                trade_type=Case(
                    When(shares__gte=0, then=Value(True)),
                    When(shares__lt=0, then=Value(False)),
                    output_field=BooleanField(),
                ),
                trade_comment=F("trade__comment"),
            )
            .select_related(
                "account",
                "trade",
                "product",
                "creator",
                "claimant",
            )
            .defer("account__reference_id")
        )


class ClaimAccountModelViewSet(ClaimModelViewSet):
    title_config_class = ClaimAccountTitleConfig
    endpoint_config_class = ClaimAccountEndpointConfig

    @cached_property
    def account(self):
        return get_object_or_404(Account, id=self.kwargs["account_id"])

    def get_queryset(self):
        return super().get_queryset().filter(account__in=self.account.get_descendants(include_self=True))


class ClaimProductModelViewSet(ClaimModelViewSet):
    title_config_class = ClaimProductTitleConfig
    endpoint_config_class = ClaimProductEndpointConfig

    @cached_property
    def product(self):
        return get_object_or_404(Product, id=self.kwargs["product_id"])

    def get_queryset(self):
        return super().get_queryset().filter(product=self.product)

    def get_identifier(self, request, identifier=None):
        return "commission:product-claim"


class ClaimEntryModelViewSet(ClaimModelViewSet):
    serializer_class = ClaimModelSerializer

    display_config_class = ClaimDisplayConfig
    title_config_class = ClaimEntryTitleConfig
    endpoint_config_class = ClaimEntryEndpointConfig
    ordering = ["id"]

    def setup(self, request, *args, **kwargs):
        super().setup(request, *args, **kwargs)
        self.kwargs["claimant_id"] = self.kwargs["entry_id"]  # ensure claimant_id is available for the serializer

    @cached_property
    def entry(self):
        return Entry.objects.get(id=self.kwargs["entry_id"])

    def get_queryset(self):
        return super().get_queryset().filter_for_customer(self.entry)


class ClaimTradeModelViewSet(ClaimModelViewSet):
    serializer_class = ClaimTradeModelSerializer

    title_config_class = ClaimTradeTitleConfig
    endpoint_config_class = ClaimTradeEndpointConfig
    button_config_class = ClaimTradeButtonConfig

    @cached_property
    def trade(self) -> Trade:
        return get_object_or_404(Trade, pk=self.kwargs["trade_id"])

    @cached_property
    def product(self) -> Product | None:
        with suppress(Product.DoesNotExist):
            return Product.objects.get(id=self.trade.underlying_instrument.id)

    def get_queryset(self):
        return super().get_queryset().filter(trade__id=self.kwargs["trade_id"])

    @action(methods=["POST"], detail=False)
    def transfer_trade(self, request, trade_id):
        serializer = TransferTradeSerializer(data=request.data)
        if serializer.is_valid():
            trade = Trade.objects.get(id=trade_id)
            transfer_date = (parse_date(serializer.data["transfer_date"]) - pd.tseries.offsets.BDay(0)).date()
            from_account = serializer.data["from_account"]
            to_account = serializer.data["to_account"]

            if not transfer_date or not from_account or not to_account:
                return Response(
                    {"__notification": {"title": "Trade not Transferred."}}, status=status.HTTP_400_BAD_REQUEST
                )

            Claim.objects.create(
                trade=trade,
                shares=trade.shares * -1,
                bank=trade.bank,
                date=transfer_date,
                product=trade.product,
                account_id=from_account,
                claimant_id=request.user.profile.id,
            )

            Claim.objects.create(
                trade=trade,
                shares=trade.shares,
                bank=trade.bank,
                date=transfer_date,
                product=trade.product,
                account_id=to_account,
                claimant_id=request.user.profile.id,
            )

            return Response({"__notification": {"title": "Trade Transferred."}}, status=status.HTTP_200_OK)
        return Response(serializer.errors, status=status.HTTP_400_BAD_REQUEST)

    @action(methods=["POST"], detail=False)
    def quick_claim(self, request, trade_id):
        trade = Trade.objects.get(id=trade_id)

        account = request.data.get("account")

        if not account:
            return Response({"__notification": {"title": "Trade not Claimed."}}, status=status.HTTP_400_BAD_REQUEST)

        Claim.objects.create(
            trade=trade,
            shares=trade.shares,
            bank=trade.bank,
            date=trade.transaction_date,
            product=trade.product,
            account_id=account,
            claimant_id=request.user.profile.id,
        )

        return Response({"__notification": {"title": "Trade Claimed."}}, status=status.HTTP_200_OK)


# Abstract AUM Viewset


class ConsolidatedTradeSummaryTableView(ClaimPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "commission:aum"

    def get_pandas_fields(self, request):
        fields = [
            pf.PKField(key="id", label="ID"),
            pf.CharField(key="title", label="Title"),
            pf.DateField(key="initial_investment_date", label="First Investment"),
            pf.SparklineField(key="aum_sparkline", label="AUM", dimension="double"),
            pf.FloatField(key="sum_shares_start", label="Shares Start", precision=0),
            pf.FloatField(key="sum_shares_end", label="Shares End", precision=0),
            pf.FloatField(key="sum_shares_diff", label="Share difference", precision=0),
            pf.FloatField(key="sum_shares_perf", label="Share Perf", precision=2, percent=True),
            pf.FloatField(
                key="sum_aum_start",
                label="AUM Start",
                precision=0,
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(
                key="sum_aum_end",
                label="AUM End",
                precision=0,
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(
                key="sum_aum_diff",
                label="Nominal difference",
                precision=0,
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(key="sum_aum_perf", label="% difference", precision=2, percent=True),
            pf.FloatField(
                key="sum_nnm_total",
                label="NNM",
                precision=0,
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
            ),
            pf.FloatField(key="sum_nnm_perf", label="NNM (%)", precision=2, percent=True),
            pf.FloatField(key="total_performance", label="Performance", precision=2),
            pf.FloatField(key="total_performance_perf", label="Performance (%)", precision=2, percent=True),
        ]
        for nnm_column in self.nnm_monthly_columns:
            fields.append(
                pf.FloatField(
                    key=nnm_column[0],
                    label=nnm_column[1],
                    precision=0,
                    decorators=(decorator(decorator_type="text", position="left", value="$"),),
                ),
            )
        if self.commission_type_columns:
            fields.append(
                pf.FloatField(
                    key="rebate_total",
                    label="Rebate",
                    precision=0,
                    decorators=(decorator(decorator_type="text", position="left", value="$"),),
                )
            )

            for ct in self.commission_type_columns:
                fields.append(
                    pf.FloatField(
                        key=ct[0],
                        label=ct[1],
                        precision=0,
                        decorators=(decorator(decorator_type="text", position="left", value="$"),),
                    )
                )

        return pf.PandasFields(fields=fields)

    queryset = Claim.objects.filter(status=Claim.Status.APPROVED, account__isnull=False)

    filterset_class = ConsolidatedTradeSummaryTableFilterSet

    search_fields = ["title"]

    def get_ordering_fields(self):
        fields = [
            "title",
            "sum_shares_start",
            "sum_shares_end",
            "sum_shares_diff",
            "sum_shares_perf",
            "sum_aum_start",
            "sum_aum_end",
            "sum_aum_diff",
            "sum_aum_perf",
            "sum_nnm_total",
            "sum_nnm_perf",
            "total_performance",
            "total_performance_perf",
            *map(lambda x: x[0], self.nnm_monthly_columns),
        ]
        if self.commission_type_columns:
            fields.extend(list(map(lambda x: x[0], self.commission_type_columns)))
            fields.append("rebate_total")
        return fields

    ordering = ["-sum_aum_end"]
    display_config_class = ConsolidatedTradeSummaryDisplayConfig
    title_config_class = ConsolidatedTradeSummaryTitleConfig
    endpoint_config_class = ConsolidatedTradeSummaryEndpointConfig
    button_config_class = ConsolidatedTradeSummaryButtonConfig

    @cached_property
    def commission_type_columns(self) -> list[tuple[str, str]]:
        if apps.is_installed("wbcommission"):
            from wbcommission.models.commission import CommissionType

            return list(map(lambda x: ("rebate_" + x[0], x[1]), CommissionType.objects.values_list("key", "name")))

    @cached_property
    def nnm_monthly_columns(self) -> list[tuple[str, str]]:
        res = []
        if self.start_date and self.end_date:
            # + MonthEnd to ensure that we include the current month
            for _d in pd.date_range(
                self.start_date, self.end_date - timedelta(days=1) + pd.offsets.MonthEnd(), freq="ME"
            ):
                res.append(("sum_nnm_" + _d.strftime("%Y-%m"), _d.strftime("%b %Y")))
        return res

    @cached_property
    def start_date(self) -> date:
        return get_date_interval_from_request(self.request, exclude_weekend=True, left_interval_inclusive=True)[0]

    @cached_property
    def end_date(self) -> date:
        return get_date_interval_from_request(self.request, exclude_weekend=True, right_interval_inclusive=True)[1]

    @cached_property
    def groupby_classification_group(self) -> ClassificationGroup:
        try:
            return ClassificationGroup.objects.get(id=self.request.GET["groupby_classification_group"])
        except (ValueError, KeyError):
            return get_default_classification_group()

    @cached_property
    def unique_product(self) -> bool:
        return self.request.GET.get("product") is not None

    def _rebate_df(self, df_aum_end: pd.Series, **kwargs) -> pd.DataFrame:
        df = pd.DataFrame()
        if apps.is_installed("wbcommission"):
            from wbcommission.viewsets.rebate import RebatePandasView

            rebate_view = RebatePandasView()
            rebate_view.request = self.request
            df = rebate_view._get_dataframe(**kwargs).drop(["title", "index"], axis=1, errors="ignore").set_index("id")
            if not df.empty:
                return df.reindex(df_aum_end.index, fill_value=0)
        return df

    def add_messages(self, request, queryset=None, paginated_queryset=None, instance=None, initial=False):
        if self.unique_product:
            return warning(
                request, "Internal trades are excluded from the Net New Money total due to the selected product."
            )

    def get_dataframe(
        self,
        request,
        queryset,
        with_rebate_df: bool = True,
        with_aum_sparkline: bool = True,
        with_neg_pos_nnm: bool = False,
        **kwargs,
    ):
        groupby = self.request.GET.get("group_by", "PRODUCT")
        groupby_map = ClaimGroupbyChoice.get_map(groupby)
        pivot = groupby_map["pk"]
        pivot_label = groupby_map["title_key"]

        cts_generator = ConsolidatedTradeSummary(
            queryset,
            self.start_date,
            self.end_date,
            pivot,
            pivot_label,
            classification_group=self.groupby_classification_group,
        )

        df = cts_generator.df
        if not df.empty:
            df_aum = cts_generator.get_aum_df()
            df_nnm = cts_generator.get_nnm_df(filter_internal_trade=not self.unique_product)
            if with_neg_pos_nnm:
                self.df_nnm_neg = cts_generator.get_nnm_df(
                    only_negative=True, filter_internal_trade=not self.unique_product
                )
                self.df_nnm_pos = cts_generator.get_nnm_df(
                    only_positive=True, filter_internal_trade=not self.unique_product
                )

            df_aum_sparkline = pd.DataFrame()
            if with_aum_sparkline:
                df_aum_sparkline = cts_generator.get_aum_sparkline()
            df_rebate = pd.DataFrame()
            if with_rebate_df:
                df_rebate = self._rebate_df(df_aum["sum_aum_end"], **kwargs)
            df_initial_investment_date = cts_generator.get_initial_investment_date_df()

            df_title = df[["id", "title"]].groupby("id").first()
            df = pd.concat(
                [df_title, df_initial_investment_date, df_aum, df_nnm, df_aum_sparkline, df_rebate], axis=1
            ).reset_index(names="id")
            df["total_performance"] = df["sum_aum_end"] - (df["sum_aum_start"] + df["sum_nnm_total"])
            df["total_performance_perf"] = df["total_performance"] / df["sum_aum_start"]

            df.loc[df["sum_aum_start"] != 0, "sum_nnm_perf"] = (
                df.loc[df["sum_aum_start"] != 0, "sum_nnm_total"] / df.loc[df["sum_aum_start"] != 0, "sum_aum_start"]
            )
            df = df[(df["sum_shares_start"].abs() > 1) | (df["sum_shares_end"].abs() > 1)]
            df = df.replace([np.inf, -np.inf, np.nan], None)
            if hasattr(self, "df_nnm_neg"):
                self.df_nnm_neg = (
                    pd.concat([df_title, self.df_nnm_neg], axis=1)
                    .reset_index()
                    .replace([np.inf, -np.inf, np.nan], None)
                    .sort_values(by="title")
                )
            if hasattr(self, "df_nnm_pos"):
                self.df_nnm_pos = (
                    pd.concat([df_title, self.df_nnm_pos], axis=1)
                    .reset_index()
                    .replace([np.inf, -np.inf, np.nan], None)
                    .sort_values(by="title")
                )
        self._columns = df.columns
        return df

    def get_aggregates(self, request, df):
        if df.empty:
            return {}
        aggregates = {}
        for col in filter(
            lambda x: ("sum" in x and "perf" not in x) or x == "total_performance" or "rebate" in x, df.columns
        ):
            aggregates[col] = {"Σ": format_number(df[col].sum())}
        return aggregates

    def get_filterset_class(self, request):
        profile = request.user.profile
        if profile.is_internal or request.user.is_superuser:
            return ConsolidatedTradeSummaryTableFilterSet
        return CustomerClaimGroupByFilter


def _sanitize_df(df: pd.DataFrame, normalization_factor: pd.Series | None = None) -> pd.DataFrame:
    df = df.replace({0: np.nan})
    scalar_columns = df.columns.difference(["title", "id"])
    if normalization_factor is not None:
        df[scalar_columns] = df[scalar_columns].div(normalization_factor.replace({0: np.nan}), axis=0)
    return df.dropna(axis=0, how="all", subset=scalar_columns)


class ConsolidatedTradeSummaryDistributionChart(ConsolidatedTradeSummaryTableView):
    WIDGET_TYPE = WidgetType.CHART.value
    IDENTIFIER = "wbportfolio:consolidatedtradesummarydistributionchart"

    title_config_class = ConsolidatedTradeSummaryDistributionChartTitleConfig
    endpoint_config_class = ConsolidatedTradeSummaryDistributionChartEndpointConfig
    button_config_class = None

    def get_filterset_class(self, request):
        return DistributionNNMChartFilter

    @cached_property
    def is_percent(self) -> bool:
        return self.request.GET.get("percent", "false") == "true"

    # TODO This is not really optimal. We need to change it at some point
    def list(self, request, *args, **kwargs):
        # we copy pasted this function from the chartviewset. Can be optimize
        figure = go.Figure()
        self.request = request
        df = self._get_dataframe(**kwargs, with_rebate_df=False, with_aum_sparkline=False, with_neg_pos_nnm=True)
        if not df.empty:
            figure = self.get_plotly(df)
        figure_json = figure.to_json()  # we serialize to use the default PlotlyEncoder
        figure_dict = json.loads(
            figure_json
        )  # we reserialize to be able to hijack the figure config. This adds an extra steps of serialization/deserialization but the overhead is negligable.
        figure_dict["config"] = {"responsive": True, "displaylogo": False}
        figure_dict["useResizeHandler"] = True
        figure_dict["style"] = {"width": "100%", "height": "100%"}
        figure_dict["messages"] = list(self._get_messages(request))

        return Response(figure_dict)

    def get_plotly(self, df):
        fig = go.Figure()
        # create the groupby NNM distribution histogram
        df = df.sort_values(by="title").set_index("id")
        normalization_factor = None
        if self.is_percent:
            normalization_factor = df["sum_aum_start"]
        nnm_monthly_columns = dict(self.nnm_monthly_columns)
        df = _sanitize_df(
            df.drop(columns=df.columns.difference(["title", *nnm_monthly_columns.keys(), "sum_nnm_total"])),
            normalization_factor=normalization_factor,
        )
        for key, label in nnm_monthly_columns.items():
            if len(nnm_monthly_columns.keys()) == 1 and hasattr(self, "df_nnm_neg") and hasattr(self, "df_nnm_pos"):
                df_nnm_pos = _sanitize_df(self.df_nnm_pos.set_index("id"), normalization_factor)
                df_nnm_neg = _sanitize_df(self.df_nnm_neg.set_index("id"), normalization_factor)
                if key in df_nnm_neg.columns:
                    fig.add_trace(
                        go.Histogram(
                            histfunc="sum",
                            y=df_nnm_neg[key],
                            x=df_nnm_neg["title"],
                            name=label + " (Negative)",
                            marker_color="#FF6961",
                        )
                    )
                if key in df_nnm_pos.columns:
                    fig.add_trace(
                        go.Histogram(
                            histfunc="sum",
                            y=df_nnm_pos[key],
                            x=df_nnm_pos["title"],
                            name=label + " (Positive)",
                            marker_color="#77DD77",
                        )
                    )
            if key in df.columns:
                figure_kwargs = {"name": label}
                if len(nnm_monthly_columns.keys()) == 1:
                    figure_kwargs["marker_color"] = "#D3D3D3"
                    figure_kwargs["name"] = label + " (Total)"
                fig.add_trace(go.Histogram(histfunc="sum", y=df[key], x=df["title"], **figure_kwargs))
        if len(nnm_monthly_columns.keys()) > 1:
            fig.add_trace(
                go.Histogram(
                    histfunc="sum",
                    y=df["sum_nnm_total"],
                    x=df["title"],
                    name=f"[{self.start_date.strftime('%Y-%m-%d')}, {self.end_date.strftime('%Y-%m-%d')}]",
                )
            )
        fig.update_layout(
            template="plotly_white",
            margin=dict(l=10, r=10, t=0, b=40),
            showlegend=True,
            legend={
                "orientation": "h",
                "yanchor": "bottom",
                "y": 1,
                "xanchor": "center",
                "x": 0.5,
            },
        )
        if self.is_percent:
            fig.update_yaxes(tickformat=".2%")
        return fig


class CumulativeNNMChartView(ConsolidatedTradeSummaryDistributionChart):
    IDENTIFIER = "wbportfolio:cumulativennmchart"

    title_config_class = CumulativeNNMChartTitleConfig
    endpoint_config_class = CumulativeNNMChartEndpointConfig
    filterset_class = CumulativeNNMChartFilter

    @cached_property
    def projected_monthly_nnm_target(self) -> int:
        try:
            return int(self.request.GET["projected_monthly_nnm_target"])
        except (ValueError, KeyError):
            return get_monthly_nnm_target()

    @cached_property
    def hide_projected_monthly_nnm_target(self) -> bool:
        try:
            return self.request.GET["hide_projected_monthly_nnm_target"] == "true"
        except (ValueError, KeyError):
            return False

    def get_filterset_class(self, request):
        return CumulativeNNMChartFilter

    def get_plotly(self, df):
        fig = go.Figure()

        # create the cumulative NNM histogram
        nnm_monthly_columns_dict = dict(self.nnm_monthly_columns)

        nnm_monthly_columns = df.columns.intersection(nnm_monthly_columns_dict.keys())
        if not nnm_monthly_columns.empty and not df.empty:
            monthly_nnm = df[[*nnm_monthly_columns]].sum().cumsum().rename(index={**nnm_monthly_columns_dict})

            fig.add_trace(
                go.Histogram(
                    histfunc="sum",
                    y=monthly_nnm,
                    x=monthly_nnm.index,
                    autobinx=False,
                    showlegend=False,
                    yaxis="y",
                    marker_color="darkgrey",
                    name="Monthly Cumulative NNM",
                )
            )
            if not self.hide_projected_monthly_nnm_target:
                a = self.projected_monthly_nnm_target
                x = np.linspace(0, monthly_nnm.shape[0], monthly_nnm.shape[0] - 1)
                target_points = a * x + a
                fig.add_trace(
                    go.Scatter(
                        x=monthly_nnm.index[0:],
                        y=target_points,
                        mode="lines",
                        text="Projected NNM target",
                        line=dict(color="red", width=4),
                        showlegend=False,
                    )
                )
        return fig


class ProfitAndLossPandasView(ClaimPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "commission:pnl"
    # LIST_DOCUMENTATION = "wbportfolio/commission/viewsets/documentation/profitandlosspandasview.md"
    # Averaging method based on https://www.tradingtechnologies.com/xtrader-help/fix-adapter-reference/pl-calculation-algorithm/understanding-pl-calculations/
    pandas_fields = pf.PandasFields(
        fields=(
            pf.PKField(key="id", label="ID"),
            pf.CharField(key="title", label="Product"),
            pf.FloatField(
                key="unrealized_pnl",
                label="Realized P&L",
                precision=2,
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
                help_text="P&L_unrealized (points) = (Theoretical Exit Price - Avg Buy Price) *  total_shares",
            ),
            pf.FloatField(
                key="realized_pnl",
                label="Unrealized P&L",
                precision=2,
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
                help_text="P&L_realized (points) = (Avg Sell Price - Avg Buy Price) * total_shares_sold",
            ),
            pf.FloatField(
                key="total_pnl",
                label="Total P&L",
                precision=2,
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
                help_text="P&L_total (points) = P&L_unrealized + P&L_realized",
            ),
            pf.FloatField(
                key="total_invested",
                label="Total AUM",
                precision=2,
                decorators=(decorator(decorator_type="text", position="left", value="$"),),
                help_text="total_invested = Avg Buy Price * total_shares",
            ),
            pf.BooleanField(
                key="is_invested",
                label="Invested",
            ),
            pf.FloatField(
                key="performance",
                label="Performance",
                precision=2,
                percent=True,
                help_text="performance = P&L_total / total_invested",
            ),
        )
    )

    queryset = Claim.objects.filter(account__owner__isnull=False)

    filterset_class = ProfitAndLossPandasFilter

    search_fields = ("title",)
    ordering_fields = ["title", "unrealized_pnl", "realized_pnl", "total_pnl", "total_invested", "performance"]

    display_config_class = ProfitAndLossPandasDisplayConfig
    title_config_class = ProfitAndLossPandasTitleConfig
    endpoint_config_class = ProfitAndLossPandasEndpointConfig

    def get_aggregates(self, request, df):
        if df.empty:
            return {}
        return {
            "unrealized_pnl": {"Σ": format_number(df.unrealized_pnl.sum())},
            "realized_pnl": {"Σ": format_number(df.realized_pnl.sum())},
            "total_pnl": {"Σ": format_number(df.total_pnl.sum())},
            "total_invested": {"Σ": format_number(df.total_invested.sum())},
        }

    def get_queryset(self):
        start_date, end_date = get_date_interval_from_request(self.request, exclude_weekend=True)
        return (
            super()
            .get_queryset()
            .annotate(
                net_value=InstrumentPrice.subquery_closest_value(
                    "net_value", instrument_pk_name="product__pk", date_lookup="exact"
                ),
                fx_rate=CurrencyFXRates.get_fx_rates_subquery("date", lookup_expr="exact"),
                total_value=ExpressionWrapper(
                    F("net_value") * F("shares") * F("fx_rate"), output_field=DecimalField()
                ),
                date_end=Least(F("product__last_valuation_date"), Value(end_date)),
                net_value_end=InstrumentPrice.subquery_closest_value(
                    "net_value", date_name="date_end", instrument_pk_name="product__pk", date_lookup="exact"
                ),
                fx_rate_end=CurrencyFXRates.get_fx_rates_subquery("date_end", lookup_expr="exact"),
                price_end=ExpressionWrapper(F("net_value_end") * F("fx_rate_end"), output_field=DecimalField()),
            )
        )

    def get_dataframe(self, request, queryset, **kwargs):
        start_date, end_date = get_date_interval_from_request(self.request, exclude_weekend=True)
        if not start_date or not end_date:
            return pd.DataFrame()

        df = pd.DataFrame(queryset.values("shares", "total_value", "price_end", "account__owner__id", "product"))
        if not df.empty:
            df[["shares", "total_value"]] = df[["shares", "total_value"]].astype("float")

            df_price_end = (
                df[["price_end", "account__owner__id", "product"]].groupby(["account__owner__id", "product"]).mean()
            )
            df_buy = df[["shares", "total_value", "account__owner__id", "product"]][df["shares"] > 0]
            df_buy = df_buy.groupby(["account__owner__id", "product"]).agg({"shares": "sum", "total_value": "sum"})
            df_buy["total_value"] = df_buy.total_value / df_buy.shares
            df_buy = df_buy.rename(columns={"total_value": "avg_buy_price", "shares": "total_buy_shares"})

            df_sell = df[["shares", "total_value", "account__owner__id", "product"]][df["shares"] < 0]
            df_sell = df_sell.groupby(["account__owner__id", "product"]).agg({"shares": "sum", "total_value": "sum"})
            df_sell = df_sell.abs()
            df_sell["total_value"] = df_sell.total_value / df_sell.shares
            df_sell = df_sell.rename(columns={"total_value": "avg_sell_price", "shares": "total_sell_shares"})

            df = pd.concat([df_buy, df_sell], axis=1).fillna(0)
            df["realized_pnl"] = (df.avg_sell_price - df.avg_buy_price) * df.total_sell_shares
            df["total_shares"] = df.total_buy_shares - df.total_sell_shares

            df["unrealized_pnl"] = (df_price_end["price_end"] - df.avg_buy_price) * df.total_shares
            df["total_pnl"] = df.unrealized_pnl + df.realized_pnl
            df = df.groupby(level=0).sum().reset_index()

            df["total_invested"] = df.avg_buy_price * df.total_buy_shares

            df["performance"] = df.total_pnl.astype(float).divide(df.total_invested.astype(float))

            df["id"] = df["account__owner__id"]
            df["title"] = df["account__owner__id"].map(
                dict(Entry.objects.filter(id__in=df["id"]).values_list("id", "computed_str"))
            )
        return df.where(pd.notnull(df), None)


class NegativeTermimalAccountPerProductModelViewSet(ClaimPermissionMixin, wb_viewsets.ReadOnlyInfiniteModelViewSet):
    IDENTIFIER = "wbportfolio:negativeaccountproduct"

    serializer_class = NegativeTermimalAccountPerProductModelSerializer
    filterset_class = NegativeTermimalAccountPerProductFilterSet
    queryset = Claim.objects.filter(account__is_active=True)

    search_fields = ("account__title", "product__computed_str")
    ordering_fields = ("account__title", "product__computed_str", "sum_shares")
    ordering = "sum_shares"

    display_config_class = NegativeTermimalAccountPerProductDisplayConfig
    title_config_class = NegativeTermimalAccountPerProductTitleConfig
    endpoint_config_class = NegativeTermimalAccountPerProductEndpointConfig

    def get_aggregates(self, queryset, paginated_queryset):
        return {
            "sum_shares": {"Σ": format_number(queryset.aggregate(s=Sum("sum_shares"))["s"] or 0)},
        }

    def get_queryset(self):
        return (
            super()
            .get_queryset()
            .select_related("account", "product")
            .values("account", "product")
            .annotate(
                sum_shares=Sum("shares"),
                id=Concat(F("account__id"), Value("-"), F("product__id"), output_field=CharField()),
                account_repr=F("account__computed_str"),
                product_repr=F("product__computed_str"),
                account_id=F("account__id"),
                product_id=F("product__id"),
            )
            .filter(sum_shares__lt=0, account__isnull=False)
            .exclude(status=Claim.Status.WITHDRAWN)
            .values("account_repr", "product_repr", "account_id", "product_id", "sum_shares", "id")
        )



from datetime import date
from io import BytesIO

import msoffcrypto
import numpy as np
import pandas as pd

from wbportfolio.models import Product, Trade


def get_portfolio(x):
    product = Product.objects.get(id=x)
    portfolio = product.primary_portfolio
    return portfolio.id


def parse(import_source):
    if credential := import_source.source.get_valid_credential(date.today()):
        password = credential.password
        buffer = BytesIO()

        encrypted_file = msoffcrypto.OfficeFile(import_source.file)
        encrypted_file.load_key(password=password)
        encrypted_file.decrypt(buffer)

        buffer.seek(0)

        df = pd.read_excel(buffer.read(), usecols=[3, 4, 7, 10, 15, 16, 17, 18, 20, 21], engine="openpyxl")

        # Filter out all non transaction rows and remove RECORD_DESC col
        df = df[df["RECORD_DESC"].str.strip() == "TRANSACTION"]
        del df["RECORD_DESC"]

        # Convert timestamps to json conform date strings
        df["TRANS_DATE"] = df["TRANS_DATE"].apply(lambda x: x.date().strftime("%Y-%m-%d"))
        df["SETTL_DATE"] = df["SETTL_DATE"].apply(lambda x: x.date().strftime("%Y-%m-%d"))

        # Replace all nan values with empty str
        df[["REGISTER_FIRSTNAME", "CUST_REF"]] = df[["REGISTER_FIRSTNAME", "CUST_REF"]].replace(np.nan, "", regex=True)

        # Merge REGISTER_FIRSTNAME and CUST_REF and then remove both cols
        df["note"] = df["REGISTER_FIRSTNAME"] + df["CUST_REF"]
        del df["REGISTER_FIRSTNAME"]
        del df["CUST_REF"]

        # Create Product Mapping and apply to df
        product_mapping = {
            product["isin"]: product["id"]
            for product in Product.objects.filter(isin__in=df["ISIN"].unique()).values("id", "isin")
        }
        df["ISIN"] = df["ISIN"].apply(lambda x: product_mapping[x])

        # Rename Columns
        df.columns = [
            "bank",
            "underlying_instrument",
            "transaction_date",
            "value_date",
            "external_id",
            "price",
            "shares",
            "comment",
        ]
        df["transaction_subtype"] = df.shares.apply(
            lambda x: Trade.Type.REDEMPTION if x < 0 else Trade.Type.SUBSCRIPTION
        )
        df["portfolio"] = df["underlying_instrument"].apply(lambda x: get_portfolio(x))
        df["underlying_instrument"] = df["underlying_instrument"].apply(
            lambda x: {"id": x, "instrument_type": "product"}
        )

        # Convert df to list of dicts
        data = list(df.T.to_dict().values())

        return {"data": data}
    return {"data": list()}



import math

import cvxpy as cp
import numpy as np

from wbportfolio.pms.typing import TradeBatch


class TradeShareOptimizer:
    def __init__(self, batch: TradeBatch, portfolio_total_value: float):
        self.batch = batch
        self.portfolio_total_value = portfolio_total_value

    def optimize(self, target_cash: float = 0.99):
        try:
            return self.optimize_trade_share(target_cash)
        except ValueError:
            return self.floor_trade_share()

    def optimize_trade_share(self, target_cash: float = 0.01):
        prices_fx_portfolio = np.array([trade.price_fx_portfolio for trade in self.batch.trades])
        target_allocs = np.array([trade.target_weight for trade in self.batch.trades])

        # Decision variable: number of shares (integers)
        shares = cp.Variable(len(prices_fx_portfolio), integer=True)

        # Calculate portfolio values
        portfolio_values = cp.multiply(shares, prices_fx_portfolio)

        # Target values based on allocations
        target_values = self.portfolio_total_value * target_allocs

        # Objective: minimize absolute deviation from target values
        objective = cp.Minimize(cp.sum(cp.abs(portfolio_values - target_values)))

        # Constraints
        constraints = [
            shares >= 0,  # No short selling
            cp.sum(portfolio_values) <= self.portfolio_total_value,  # Don't exceed budget
            cp.sum(portfolio_values) >= (1.0 - target_cash) * self.portfolio_total_value,  # Use at least 99% of budget
        ]

        # Solve
        problem = cp.Problem(objective, constraints)
        problem.solve(solver=cp.CBC)

        if problem != "optimal":
            raise ValueError(f"Optimization failed: {problem.status}")

        shares_result = shares.value.astype(int)
        return TradeBatch(
            [
                trade.normalize_target(target_shares=shares_result[index])
                for index, trade in enumerate(self.batch.trades)
            ]
        )

    def floor_trade_share(self):
        return TradeBatch(
            [trade.normalize_target(target_shares=math.floor(trade.target_shares)) for trade in self.batch.trades]
        )



from contextlib import suppress
from datetime import date, datetime, time, timedelta
from decimal import Decimal

from celery import shared_task
from django.contrib import admin
from django.contrib.contenttypes.fields import GenericRelation
from django.contrib.postgres.constraints import ExclusionConstraint
from django.contrib.postgres.fields import DateRangeField, RangeOperators
from django.db import models
from django.db.models import (
    BooleanField,
    Case,
    DecimalField,
    ExpressionWrapper,
    F,
    OuterRef,
    Subquery,
    Value,
    When,
)
from django.db.models.functions import Coalesce
from django.db.models.signals import pre_save
from django.dispatch import receiver
from pandas.tseries.offsets import BDay
from wbcore.contrib.ai.llm.config import add_llm_prompt
from wbcore.contrib.currency.models import CurrencyFXRates
from wbcore.contrib.directory.models import Entry
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.contrib.notifications.utils import create_notification_type
from wbcore.permissions.shortcuts import get_internal_users
from wbcore.signals import pre_merge
from wbcore.utils.enum import ChoiceEnum
from wbcore.workers import Queue
from wbcrm.models.accounts import Account
from wbfdm.models.instruments.instrument_prices import InstrumentPrice
from wbfdm.models.instruments.instruments import InstrumentManager, InstrumentType
from wbreport.models import Report

from wbportfolio.models.portfolio_relationship import InstrumentPortfolioThroughModel

from ..preferences import get_product_termination_notice_period
from . import PortfolioRole
from .llm.wbcrm.analyze_relationship import get_holding_prompt
from .mixins.instruments import PMSInstrument, PMSInstrumentAbstractModel


class TypeOfReturn(ChoiceEnum):
    TOTAL_RETURN = "Total return"
    YIELD = "Yield"
    PRICE_RETURN = "Price return"
    ABSOLUT_RETURN = "Absolut return"


class AssetClass(ChoiceEnum):
    EQUITY = "Equity"
    FIXED_INCOME = "Fixed income"
    PRIVATE_EQUITY = "Private equity"
    DERIVATIVE = "Derivative"
    CASH = "Cash"
    CRYPTOCURRENCY = "Cryptocurrency"
    COMMODITY = "Commodity"
    HEDGE_FUND = "Hedge Fund"
    MULTI_ASSET = "Multi Asset"


class LegalStructure(ChoiceEnum):
    AMC = "Actively Managed Certificate"
    SICAV_UCITS = "SICAV (UCITS)"
    PLC_UCITS = "PLC (UCITS)"
    SICAV_AIF = "SICAV (AIFs)"
    SCF_UCITS = "Swiss Contractual Fund (UCITS)"
    FCP_AIF = "FCP (AIFs)"
    PLC_AIF = "PLC (AIFs)"
    SCF_AIF = "Swiss Contractual Fund (AIF) "


class InvestmentIndex(ChoiceEnum):
    LONG = "Long"
    SHORT = "Short"
    LONG_SHORT = "Long/Short"


class Liquidy(ChoiceEnum):
    DAILY = "Daily"
    WEEKLY = "Weekly"
    MONTHLY = "Monthly"
    QUATERLY = "Quaterly"
    YEARLY = "Yearly"


class DefaultProductManager(InstrumentManager):
    def get_queryset(self):
        today = date.today()
        return (
            super()
            .get_queryset()
            .annotate(
                value_date=Case(
                    When(last_valuation_date__isnull=False, then=F("last_valuation_date")),
                    default=Value(today),
                    output_field=models.DateField(),
                ),
                is_invested=models.Exists(
                    InstrumentPortfolioThroughModel.objects.filter(
                        instrument=models.OuterRef("pk"),
                        portfolio__invested_timespan__startswith__lte=OuterRef("value_date"),
                        portfolio__invested_timespan__endswith__gt=OuterRef("value_date"),
                    )
                ),
                current_bank_fees=Coalesce(
                    Subquery(
                        FeeProductPercentage.objects.filter(
                            product=OuterRef("pk"),
                            type=FeeProductPercentage.Type.BANK,
                            timespan__startswith__lte=OuterRef("value_date"),
                            timespan__endswith__gt=OuterRef("value_date"),
                        ).values("percent")[:1]
                    ),
                    Decimal(0),
                ),
                current_management_fees=Coalesce(
                    Subquery(
                        FeeProductPercentage.objects.filter(
                            product=OuterRef("pk"),
                            type=FeeProductPercentage.Type.MANAGEMENT,
                            timespan__startswith__lte=OuterRef("value_date"),
                            timespan__endswith__gt=OuterRef("value_date"),
                        ).values("percent")[:1]
                    ),
                    Decimal(0),
                ),
                current_performance_fees=Coalesce(
                    Subquery(
                        FeeProductPercentage.objects.filter(
                            product=OuterRef("pk"),
                            type=FeeProductPercentage.Type.PERFORMANCE,
                            timespan__startswith__lte=OuterRef("value_date"),
                            timespan__endswith__gt=OuterRef("value_date"),
                        ).values("percent")[:1]
                    ),
                    Decimal(0),
                ),
                current_performance_fees_vat_deduction=Coalesce(
                    Subquery(
                        FeeProductPercentage.objects.filter(
                            product=OuterRef("pk"),
                            type=FeeProductPercentage.Type.PERFORMANCE,
                            timespan__startswith__lte=OuterRef("value_date"),
                            timespan__endswith__gt=OuterRef("value_date"),
                        ).values("vat_deduction")[:1]
                    ),
                    Decimal(0),
                ),
                current_gross_performance_fees_percent=F("current_performance_fees")
                + F("current_performance_fees_vat_deduction"),
                current_total_issuer_fees=F("current_bank_fees") + F("current_management_fees"),
            )
        )


class FeeProductPercentage(models.Model):
    class Type(models.TextChoices):
        MANAGEMENT = "MANAGEMENT", "Management"
        PERFORMANCE = "PERFORMANCE", "Performance"
        BANK = "BANK", "Bank"

    product = models.ForeignKey("wbportfolio.Product", related_name="fees_percentages", on_delete=models.CASCADE)
    type = models.CharField(max_length=16, default=Type.MANAGEMENT, choices=Type.choices)
    vat_deduction = models.DecimalField(
        max_digits=6,
        decimal_places=6,
        default=Decimal(0),
        verbose_name="VAT Deduction",
        help_text="The VAT deducted from this fees percentage",
    )
    percent = models.DecimalField(max_digits=6, decimal_places=6)
    timespan = DateRangeField()

    class Meta:
        verbose_name = "Fees Percentage"
        verbose_name_plural = "Fees Percentage"
        constraints = [
            ExclusionConstraint(
                name="exclude_overlapping_product_fees_type",
                expressions=[
                    ("timespan", RangeOperators.OVERLAPS),
                    ("product", RangeOperators.EQUAL),
                    ("type", RangeOperators.EQUAL),
                ],
            ),
        ]

    def __str__(self) -> str:
        return f"{self.product.name} ({self.type})"

    @property
    def net_percent(self) -> Decimal:
        return self.percent

    @property
    def gross_percent(self) -> Decimal:
        return self.percent + self.vat_deduction


class Product(PMSInstrumentAbstractModel):
    reports = GenericRelation(Report)

    share_price = models.PositiveIntegerField(
        default=100,
        verbose_name="Share Price",
        help_text="The initial share price that is used to calculate the nominal value of a product.",
    )
    initial_high_water_mark = models.PositiveIntegerField(
        default=100,
        verbose_name="Initial High Water Mark",
        help_text="Initial High Water Mark",
    )

    bank = models.ForeignKey(
        "directory.Company",
        related_name="issues_products",
        on_delete=models.PROTECT,
        verbose_name="Bank",
        help_text="The Bank that holds the product. A company from the CRM.",
    )

    white_label_customers = models.ManyToManyField(
        "directory.Entry",
        related_name="white_label_customer_products",
        blank=True,
        verbose_name="White Label Customers",
        help_text="Specifies whether a product is a white label product or not. If at least one customer is specified, this product becomes a white label product and can only be seen by the respected customers.",
    )

    default_account = models.ForeignKey(
        "wbcrm.Account",
        related_name="default_product_accounts",
        limit_choices_to=models.Q(is_terminal_account=True),
        blank=True,
        null=True,
        on_delete=models.SET_NULL,
        verbose_name="Default Account",
        help_text="If a default Account is set, then all newly created tradesare automatically matched to this sub account.",
    )

    fee_calculation = models.ForeignKey(
        "wbportfolio.FeeCalculation",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        related_name="products",
        verbose_name="Fee Calculation Method",
    )

    termsheet = models.FileField(max_length=256, null=True, blank=True, upload_to="portfolio/product/termsheets")

    class Dividend(models.TextChoices):
        CAPITALISATION = "CAPITALISATION", "Capitalisation"
        DISTRIBUTION = "DISTRIBUTION", "Distribution"

    type_of_return = models.CharField(
        max_length=16,
        default=TypeOfReturn.TOTAL_RETURN.name,
        choices=TypeOfReturn.choices(),
        verbose_name="Type of Return",
    )
    asset_class = models.CharField(
        max_length=16,
        default=AssetClass.EQUITY.name,
        choices=AssetClass.choices(),
        verbose_name="Asset Class",
    )
    legal_structure = models.CharField(
        max_length=16,
        default=LegalStructure.AMC.name,
        choices=LegalStructure.choices(),
        verbose_name="Legal Structure",
    )
    investment_index = models.CharField(
        max_length=16,
        default=InvestmentIndex.LONG.name,
        choices=InvestmentIndex.choices(),
        verbose_name="Long/Short Strategy",
    )
    liquidity = models.CharField(
        max_length=16,
        default=Liquidy.DAILY.name,
        choices=Liquidy.choices(),
        verbose_name="Liquidity",
    )

    jurisdiction = models.ForeignKey(
        "geography.Geography",
        related_name="products",
        null=True,
        blank=True,
        on_delete=models.SET_NULL,
        verbose_name="Country on which this product is issued and bonded by law",
        help_text="The country of jurisdiction",
        limit_choices_to={"level": 1},
    )

    dividend = models.CharField(
        max_length=16,
        default=Dividend.CAPITALISATION,
        choices=Dividend.choices,
        verbose_name="Type of Dividends",
    )

    minimum_subscription = models.IntegerField(
        default=100,
        help_text="Minimum subscription amount allowed to invest in the fund",
    )

    cut_off_time = models.TimeField(default=time(13, 0, 0))
    external_webpage = models.URLField(blank=True, null=True)

    bank_account = models.ForeignKey(
        to="directory.BankingContact",
        related_name="wbportfolio_products",
        null=True,
        blank=True,
        on_delete=models.PROTECT,
    )

    objects = DefaultProductManager()

    class Meta:
        verbose_name = "Product"
        verbose_name_plural = "Products"
        notification_types = [
            create_notification_type(
                "wbportfolio.product.termination_notice",
                "Product Termination Notice",
                "Sends a notification when a product is expected termination in the near future",
                True,
                True,
                True,
                is_lock=True,
            ),
        ]

    def pre_save(self):
        super().pre_save()
        self.instrument_type = InstrumentType.PRODUCT
        if "market_data" not in self.dl_parameters:
            # we default to the internal dataloader
            self.dl_parameters["market_data"] = {
                "path": "wbfdm.contrib.internal.dataloaders.market_data.MarketDataDataloader"
            }

        self.is_managed = True

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)
        if self.delisted_date and self.delisted_date <= date.today():
            self.reports.update(is_active=False)

    def get_title(self):
        if self.parent:
            return f"{self.parent.name} ({self.name})"
        return self.name

    def compute_str(self):
        computed_str = f"{self.get_title()} - {self.bank.name} - {self.currency.key}"
        if self.id:
            fees = (self.bank_fees + self.management_fees) * 100
            computed_str += f" - {fees:.2f} % Mgmt Fees"
        computed_str += f" ({self.isin})"
        return computed_str

    def check_and_notify_product_termination_on_date(self, today: date) -> bool:
        """
        Checks if today is the expected notice date for product termination and sends notifications if applicable.

        The expected notice date is calculated by subtracting the product termination notice period from the product's delisted date. Notifications are sent to users associated with the product's portfolio roles.

        Args:
            today (date): The date to check against the expected notice date.

        Returns:
            bool: True if termination is due, False otherwise.
        """
        if self.delisted_date:
            product_termination_expected_notice_date = self.delisted_date - timedelta(
                days=get_product_termination_notice_period()
            )
            if today == product_termination_expected_notice_date:
                roles = PortfolioRole.objects.filter(
                    (models.Q(instrument=self) | models.Q(instrument__isnull=True))
                    & (models.Q(start__isnull=True) | models.Q(start__gte=today))
                    & (models.Q(end__isnull=True) | models.Q(end__lte=today))
                    & models.Q(person__user_account__isnull=False)
                )
                for user in get_internal_users().filter(is_active=True, id__in=roles.values("person__user_account")):
                    send_notification(
                        code="wbportfolio.product.termination_notice",
                        title="Product Termination Notice",
                        body=f"The product {self} will be terminated on the {self.delisted_date:%Y-%m-%d}",
                        user=user,
                    )
                return True
        return False

    @property
    def white_label_product(self):
        return self.white_label_customers.all().count() > 0

    @property
    def urlify_title(self):
        url = self.name + "-" + self.bank.name + "-" + self.isin + "-" + self.currency.key
        return url.lower()

    @property
    @admin.display(description="Invested")
    def _is_invested(self) -> bool:
        """
        Return True if the associated asset portfolio is invested
        """
        return getattr(self, "is_invested", self.portfolio and self.portfolio.is_invested_at_date(date.today()))

    @property
    def group(self):  # for backward compatibility
        from wbportfolio.models.product_groups import ProductGroup

        if self.parent:
            with suppress(ProductGroup.DoesNotExist):
                return ProductGroup.objects.get(id=self.parent.id)
        return None

    def get_fees_percent(self, val_date: date, fee_type: FeeProductPercentage.Type, net: bool = True) -> Decimal:
        try:
            fee = FeeProductPercentage.objects.get(
                product=self, type=fee_type, timespan__startswith__lte=val_date, timespan__endswith__gt=val_date
            )
            percent = fee.percent
            if not net:
                percent += fee.vat_deduction
            return percent
        except FeeProductPercentage.DoesNotExist:
            return Decimal(0)

    @property
    def bank_fees(self) -> Decimal:
        return self.get_fees_percent(date.today(), FeeProductPercentage.Type.BANK)

    @property
    def management_fees(self) -> Decimal:
        return self.get_fees_percent(date.today(), FeeProductPercentage.Type.MANAGEMENT)

    @property
    def performance_fees(self) -> Decimal:
        return self.get_fees_percent(date.today(), FeeProductPercentage.Type.PERFORMANCE)

    @property
    def gross_performance_fees(self) -> Decimal:
        return self.get_fees_percent(date.today(), FeeProductPercentage.Type.PERFORMANCE, net=False)

    def get_high_water_mark(self, val_date: date) -> Decimal:
        """Returns the high water mark of a instrument

        The high water mark is the highest previously reached net price

        Returns:
            decimal.Decimal -- High Water Mark
        """
        latest_high_water_mark = (
            self.valuations.filter(date__lt=val_date).aggregate(high_water_mark=models.Max(F("net_value")))[
                "high_water_mark"
            ]
            or self.share_price
        )
        return max(latest_high_water_mark, self.initial_high_water_mark)

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:product"

    @classmethod
    def get_representation_endpoint(cls):
        return "wbportfolio:productrepresentation-list"

    @classmethod
    def get_representation_value_key(cls):
        return "id"

    @classmethod
    def get_representation_label_key(cls):
        return "{{computed_str}} ({{bank_name}})"

    @classmethod
    def get_last_valid_active_product_date(cls):
        current_date = (datetime.today() - BDay(1)).date()
        filter_date = current_date
        for product in Product.active_objects.all():
            if product.valuations.filter(date__lte=current_date).exists():
                last_valid_date = product.valuations.filter(date__lte=current_date).latest("date").date
                if last_valid_date < filter_date:
                    filter_date = last_valid_date
        return filter_date

    @classmethod
    def get_products(cls, person, base_qs: models.QuerySet | None = None):
        if base_qs is None:
            base_qs = cls.objects.all()
        if person.is_internal or person.user_account.is_superuser:
            return base_qs.all()
        return base_qs.filter(
            models.Q(white_label_customers__isnull=True)
            | models.Q(white_label_customers__in=[person, *person.employers.all()])
        ).distinct()

    @classmethod
    def subquery_is_white_label_product(cls, product_pk_name="pk"):
        return Coalesce(
            models.Subquery(
                Entry.objects.filter(white_label_customer_products__pk=models.OuterRef(product_pk_name))
                .values("white_label_customer_products")
                .annotate(
                    num_white_label_customers=models.Count("*"),
                    is_white_label_product=models.Case(
                        models.When(num_white_label_customers__gt=0, then=True),
                        output_field=models.BooleanField(),
                    ),
                )
                .values("is_white_label_product")[:1]
            ),
            False,
            output_field=models.BooleanField(),
        )

    @classmethod
    def annotate_last_aum(cls, qs, val_date=None, date_key="last_valuation_date"):
        if val_date:
            fx_rate_subquery = CurrencyFXRates.get_fx_rates_subquery(
                val_date, currency="currency", lookup_expr="exact"
            )
        else:
            fx_rate_subquery = CurrencyFXRates.get_fx_rates_subquery(
                date_key, currency="currency", lookup_expr="exact"
            )

        qs = InstrumentPrice.annotate_sum_shares(qs, val_date, date_key=date_key).annotate(
            net_value=Subquery(
                InstrumentPrice.objects.filter(
                    calculated=False, instrument=OuterRef("pk"), date=OuterRef("last_valuation_date")
                ).values("net_value")[:1]
            ),
            assets_under_management=Coalesce(F("sum_shares") * F("net_value"), Decimal(0)),
            fx_rate=Coalesce(
                fx_rate_subquery,
                Decimal(1.0),
            ),
            assets_under_management_usd=ExpressionWrapper(
                F("assets_under_management") * F("fx_rate"),
                output_field=DecimalField(),
            ),
            is_white_label=Case(
                When(white_label_customers__isnull=True, then=Value(False)),
                default=Value(True),
                output_field=BooleanField(),
            ),
        )
        return qs.distinct()


@shared_task(queue=Queue.DEFAULT.value)
def update_outstanding_shares_as_task(instrument_id: int, clear: bool = False):
    instrument = PMSInstrument.objects.get(id=instrument_id)
    instrument.update_outstanding_shares(clear=clear)


@receiver(pre_merge, sender="wbcrm.Account")
def handle_pre_merge_account_for_product(sender: models.Model, merged_object: Account, main_object: Account, **kwargs):
    """
    Simply reassign the product's default account linked to the merged account to the main account
    """
    Product.objects.filter(default_account=merged_object).update(default_account=main_object)


@receiver(pre_save, sender="wbfdm.Instrument")
def pre_save_instrument(sender, instance, raw, **kwargs):
    if not raw and instance.id:
        # Remove duplicates if existings
        instance.old_isins = list(set(instance.old_isins))
        pre_instance = sender.objects.get(id=instance.id)
        if (
            pre_instance.isin
            and instance.isin
            and pre_instance.isin != instance.isin
            and pre_instance.isin not in instance.old_isins
        ):
            instance.old_isins = [*instance.old_isins, pre_instance.isin]
        if pre_instance.currency != instance.currency:
            # currency has changed we need to recompute currency_fx_rate
            instance.prices.annotate(
                new_currency_fx_rate_to_usd_id=models.Subquery(
                    CurrencyFXRates.objects.filter(currency=instance.currency, date=models.OuterRef("date")).values(
                        "id"
                    )[:1]
                )
            ).update(currency_fx_rate_to_usd=models.F("new_currency_fx_rate_to_usd_id"))
            instance.assets.annotate(
                new_currency_fx_rate_instrument_to_usd_id=models.Subquery(
                    CurrencyFXRates.objects.filter(currency=instance.currency, date=models.OuterRef("date")).values(
                        "id"
                    )[:1]
                )
            ).update(currency_fx_rate_instrument_to_usd=models.F("new_currency_fx_rate_instrument_to_usd_id"))


@receiver(add_llm_prompt, sender="wbcrm.Account")
def add_performance_to_account_heat(sender, instance, key, **kwargs):
    if key == "analyze_relationship":
        return get_holding_prompt(instance)
    return []



# Generated by Django 4.2.11 on 2024-03-14 15:54

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("wbfdm", "0013_instrument_is_investable_universe_and_more"),
        ("wbportfolio", "0056_remove_companyportfoliodata_assets_under_management_currency_and_more"),
    ]

    operations = [
        migrations.AlterField(
            model_name="portfolio",
            name="preferred_instrument_classifications",
            field=models.ManyToManyField(
                blank=True,
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True), ("level", 0)),
                related_name="preferred_portfolio_classifications",
                through="wbportfolio.PortfolioInstrumentPreferredClassificationThroughModel",
                to="wbfdm.instrument",
                verbose_name="The Preferred classification per instrument",
            ),
        ),
        migrations.AlterField(
            model_name="portfolioinstrumentpreferredclassificationthroughmodel",
            name="instrument",
            field=models.ForeignKey(
                limit_choices_to=models.Q(("instrument_type__is_classifiable", True), ("level", 0)),
                on_delete=django.db.models.deletion.CASCADE,
                related_name="preferred_classification_portfolio",
                to="wbfdm.instrument",
            ),
        ),
    ]



from decimal import Decimal

import pytest
from pandas.tseries.offsets import BDay

from wbportfolio.models import AssetPosition, PortfolioRole


@pytest.mark.django_db
class TestAssetPositionModel:
    def test_init(self, asset_position):
        assert asset_position.id is not None

    def test_currency_group_by(self, asset_position_factory, equity_factory, portfolio, currency):
        asset_position_factory.create(
            portfolio=portfolio, currency=currency, underlying_instrument=equity_factory.create()
        )
        asset_position_factory.create(
            portfolio=portfolio, currency=currency, underlying_instrument=equity_factory.create()
        )
        assert (
            AssetPosition.currency_group_by(AssetPosition.objects.all()).values("groupby_id").distinct().count() == 1
        )

    def test_country_group_by(self, asset_position_factory, portfolio, equity):
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=equity)
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=equity)
        assert AssetPosition.country_group_by(AssetPosition.objects.all()).values("groupby_id").distinct().count() == 1

    def test_exchange_group_by(self, asset_position_factory, portfolio, exchange_factory, instrument):
        asset_position_factory.create(
            portfolio=portfolio, exchange=exchange_factory.create(), underlying_quote=instrument
        )
        asset_position_factory.create(
            portfolio=portfolio, exchange=exchange_factory.create(), underlying_quote=instrument
        )
        assert (
            AssetPosition.exchange_group_by(AssetPosition.objects.all()).values("groupby_id").distinct().count() == 1
        )

    def test_cash_group_by(self, asset_position_factory, portfolio, cash, equity):
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=cash)
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=cash)
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=equity)
        assert AssetPosition.cash_group_by(AssetPosition.objects.all()).values("groupby_id").distinct().count() == 2

    def test_equity_group_by(self, asset_position_factory, portfolio, equity):
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=equity)
        asset_position_factory.create(portfolio=portfolio, underlying_instrument=equity)
        assert AssetPosition.equity_group_by(AssetPosition.objects.all()).values("groupby_id").distinct().count() == 1

    def test_marketcap_group_by(
        self, asset_position_factory, equity_factory, portfolio, instrument_price_factory, currency_fx_rates_factory
    ):
        fx_rate = currency_fx_rates_factory.create(value=1)
        i1 = instrument_price_factory.create(instrument=equity_factory.create(), market_capitalization=1000000000)
        i2 = instrument_price_factory.create(instrument=equity_factory.create(), market_capitalization=30000000000)
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_quote_price=i1,
            currency_fx_rate_instrument_to_usd=fx_rate,
            currency_fx_rate_portfolio_to_usd=fx_rate,
        )  # small
        asset_position_factory.create(
            portfolio=portfolio,
            underlying_quote_price=i2,
            currency_fx_rate_instrument_to_usd=fx_rate,
            currency_fx_rate_portfolio_to_usd=fx_rate,
        )  # larg

        qs = AssetPosition.marketcap_group_by(AssetPosition.objects.all())
        assert qs.values("aggregated_title").distinct().count() == 2
        assert set(qs.distinct().values_list("aggregated_title", flat=True)) == {"10B to 50B", "< 2B"}

    def test_liquidity_group_by(self, asset_position_factory, equity_factory, portfolio, instrument_price_factory):
        a1 = asset_position_factory.create(
            portfolio=portfolio, initial_shares=1, underlying_instrument=equity_factory.create()
        )  # small
        a2 = asset_position_factory.create(
            portfolio=portfolio, initial_shares=1, underlying_instrument=equity_factory.create()
        )  # larg

        instrument_price_factory.create(instrument=a1.underlying_instrument, date=a1.date, volume_50d=1000000)
        instrument_price_factory.create(instrument=a2.underlying_instrument, date=a2.date, volume_50d=3000000)

        qs = AssetPosition.liquidity_group_by(AssetPosition.objects.all())
        assert qs.values("aggregated_title").distinct().count() == 1

    def test_get_shown_positions_superuser(self, portfolio, superuser, asset_position_factory):
        superuser.is_superuser = True
        superuser.save()
        asset_position_factory.create_batch(10, portfolio=portfolio)
        qs = AssetPosition.get_shown_positions(superuser.profile)
        assert qs.count() == 10

    def test_get_shown_positions_analyst(
        self, portfolio_factory, product_factory, person, user, asset_position_factory, product_portfolio_role_factory
    ):
        user.profile = person
        user.save()
        p2 = portfolio_factory.create()
        product = product_factory.create()
        product_portfolio_role_factory.create(
            person=person, instrument=product, role_type=PortfolioRole.RoleType.ANALYST
        )
        asset_position_factory.create_batch(10, portfolio=product.portfolio)
        asset_position_factory.create_batch(10, portfolio=p2)
        qs = AssetPosition.get_shown_positions(person)
        assert qs.count() == 10

    def test_analytical_objects(self, asset_position_factory, portfolio, equity_factory):
        # Data creation
        equity1 = equity_factory.create()
        equity2 = equity_factory.create()
        asset_position_factory.create_batch(5, underlying_instrument=equity1, portfolio=portfolio)
        asset_position_factory.create_batch(5, underlying_instrument=equity2, portfolio=portfolio)

        # We test calculation with equity1. equity2 allows to certify the temporal continuity of the portfolio.
        qs_assets_equity1 = AssetPosition.analytical_objects.order_by("date").filter(underlying_instrument=equity1)
        qs_prices_equity1 = equity1.prices.order_by("date")

        # ap holds for AssetPosition and ep holds for Equity Price.
        ap1, ap2 = qs_assets_equity1[:2]
        ep1, ep2, ep3, ep4 = qs_prices_equity1[:4]

        # Basic Performance
        assert ap2.performance == float(ap2.price_fx_usd / ap1.price_fx_usd - 1)
        assert ap2.performance == pytest.approx(float(ep2._net_value_usd / ep1._net_value_usd - Decimal(1)))

        # We sell the equity (ap3 deleted) and then we buy it again (back in ap4).
        # This is like having a missing data in between.
        qs_assets_equity1.exclude(id__in=[ap1.id, ap2.id]).first().delete()  # ap3
        ap4 = qs_assets_equity1.exclude(id__in=[ap1.id, ap2.id]).first()

        assert ap4.performance is None
        assert ep4._net_value_usd / ep3._net_value_usd is not None

    def test_get_underlying_quote_price(self, asset_position_factory, instrument_price_factory):
        """
        In this test, we confirm that an asset position with no matching real price for the same date and underlying instrument will result in an empty underlying_quote_price
        """
        a1 = asset_position_factory.create(underlying_quote_price=None)
        a1.save()
        assert a1.underlying_quote_price is None
        p2_calculated = instrument_price_factory.create(calculated=True)
        a2 = asset_position_factory.create(
            underlying_quote_price=None, underlying_instrument=p2_calculated.instrument, date=p2_calculated.date
        )
        a2.save()
        assert a2.underlying_quote_price is None
        p2_real = instrument_price_factory.create(
            calculated=False, date=p2_calculated.date, instrument=p2_calculated.instrument
        )
        a2.save()
        assert a2.underlying_quote_price == p2_real

    def test_change_currency_recompute_currency_fx_rate_for_price_and_assets(
        self, weekday, instrument, currency_fx_rates_factory, instrument_price_factory, asset_position_factory
    ):
        fx_rate = currency_fx_rates_factory.create(currency=instrument.currency, date=weekday)
        fx_rate_other = currency_fx_rates_factory.create(date=weekday)

        p = instrument_price_factory.create(instrument=instrument, date=weekday)
        a = asset_position_factory.create(underlying_instrument=instrument, date=weekday)

        assert p.currency_fx_rate_to_usd == fx_rate
        assert a.currency_fx_rate_instrument_to_usd == fx_rate

        instrument.currency = fx_rate_other.currency
        instrument.save()

        p.refresh_from_db()
        a.refresh_from_db()

        assert p.currency_fx_rate_to_usd == fx_rate_other
        assert a.currency_fx_rate_instrument_to_usd == fx_rate_other

    def test_create_price_set_assetposition(self, asset_position_factory, instrument_price_factory):
        p0 = instrument_price_factory.create()
        d1 = (p0.date + BDay(1)).date()
        a1 = asset_position_factory.create(date=d1, underlying_instrument=p0.instrument, underlying_quote_price=None)

        # check it is linked to the only price
        assert a1.underlying_quote_price == p0
        p1 = instrument_price_factory.create(instrument=p0.instrument, date=d1)
        a1.refresh_from_db()
        assert a1.underlying_quote_price == p1

        d_1 = (p0.date - BDay(1)).date()
        a_1 = asset_position_factory.create(date=d_1, underlying_instrument=p1.instrument, underlying_quote_price=None)
        assert a_1.underlying_quote_price is None
        p_1 = instrument_price_factory.create(instrument=p1.instrument, date=d_1)
        a_1.refresh_from_db()
        assert a_1.underlying_quote_price == p_1

    def test_save_asset_without_instrument(self, asset_position_factory, instrument_factory):
        parent = instrument_factory.create()
        instrument_factory.create(parent=parent)  # noise
        primary_quote = instrument_factory.create(parent=parent, is_primary=True)

        a = asset_position_factory.create(underlying_quote=primary_quote, underlying_instrument=None)
        assert a.underlying_instrument == parent

        a = asset_position_factory.create(underlying_quote=None, underlying_instrument=parent)
        assert a.underlying_quote == primary_quote



import re

from django.utils.dateparse import parse_date

from wbportfolio.models import Product

INSTRUMENT_MAP_NAME = {"EDA23_AtonRa Z class": "LU2170995018"}


def _get_exchange_from_ticker(ticker):
    try:
        ticker, exchange = ticker.split(" ")
    except Exception:
        ticker, exchange = ticker, None
    return exchange


def _get_ticker(ticker):
    return ticker.split(" ")[0]


def _get_underlying_instrument(bbg_code, name, currency, instrument_type="equity", isin=None, cash_position=False):
    if isin := INSTRUMENT_MAP_NAME.get(bbg_code, None):
        return {"isin": isin}

    exchange = _get_exchange_from_ticker(bbg_code)
    ticker = _get_ticker(bbg_code)

    cash_position = cash_position or "CASH" == bbg_code or len(re.findall("(CASH [A-Z]{3})", bbg_code)) > 0
    if cash_position:
        underlying_quote = {"instrument_type": "cash", "currency__key": currency}
    else:
        underlying_quote = {
            "exchange": {"bbg_exchange_codes": exchange},
            "currency__key": currency,
            "name": name.split(" @")[0].split("/")[
                0
            ],  # we remove anything after @ and also any trailing symbol (which represents some country information at natixis)
            "ticker": ticker,
            "instrument_type": instrument_type,
        }
        if not isin:
            isin_re = re.findall("([A-Z]{2}[A-Z0-9]{9}[0-9]{1})", ticker)
            if len(isin_re) > 0:
                isin = isin_re[0]
                # Natixis gives us ISIN as ticker for product. in that case, we registered the isin but we remove the ticker
                underlying_quote["instrument_type"] = "product"
                del underlying_quote["ticker"]
            elif Product.objects.filter(ticker=ticker).exists():
                underlying_quote["instrument_type"] = "product"
        if isin:
            underlying_quote["isin"] = isin
    return underlying_quote


def file_name_parse_isin(file_name):
    dates = re.findall(r"_([0-9]{4}[-_]?[0-9]{2}[-_]?[0-9]{2})", file_name)
    isin = re.findall(r"([A-Z]{2}(?![A-Z]{10}\b)[A-Z0-9]{10})_", file_name)
    ticker = re.findall(r"(NX[A-Z]*)_", file_name)
    if len(dates) == 0:
        raise ValueError("Not dates found in the filename")
    res = {"valuation_date": parse_date(dates[0].replace("_", "-"))}
    if len(isin) >= 1:
        res["product"] = {"isin": isin[0]}
    elif len(ticker) == 1:
        res["product"] = {"ticker": ticker[0]}
    return res



from contextlib import suppress
from datetime import date, datetime

import pandas as pd
from django.contrib.messages import warning
from django.db.models import Q
from django.utils.functional import cached_property
from wbcore.contrib.io.viewsets import ExportPandasAPIViewSet
from wbcore.contrib.pandas import fields as pf
from wbcore.utils.strings import format_number
from wbfdm.analysis.esg.enums import AggregationMethod, ESGAggregation
from wbfdm.analysis.esg.esg_analysis import DataLoader
from wbfdm.models import Instrument

from wbportfolio.filters import ESGMetricAggregationPortfolioPandasFilterSet
from wbportfolio.models import AssetPosition
from wbportfolio.viewsets.configs import (
    ESGMetricAggregationPortfolioPandasDisplayConfig,
    ESGMetricAggregationPortfolioPandasEndpointConfig,
    ESGMetricAggregationPortfolioPandasTitleConfig,
)

from .mixins import UserPortfolioRequestPermissionMixin


class ESGMetricAggregationPortfolioPandasViewSet(UserPortfolioRequestPermissionMixin, ExportPandasAPIViewSet):
    IDENTIFIER = "wbportfolio:esgmetricportfolio"
    filterset_class = ESGMetricAggregationPortfolioPandasFilterSet
    queryset = AssetPosition.objects.all()

    display_config_class = ESGMetricAggregationPortfolioPandasDisplayConfig
    title_config_class = ESGMetricAggregationPortfolioPandasTitleConfig
    endpoint_config_class = ESGMetricAggregationPortfolioPandasEndpointConfig

    search_fields = ["underlying_instrument_repr"]

    @cached_property
    def val_date(self) -> date:
        if validity_date_repr := self.request.GET.get("date"):
            val_date = datetime.strptime(validity_date_repr, "%Y-%m-%d")
        else:
            val_date = date.today()

        qs = AssetPosition.objects.filter(portfolio=self.portfolio, date__lte=val_date)
        if qs.exists():
            return qs.latest("date").date
        else:
            warning(
                self.request,
                f"No position were found before the {val_date:%Y-%m-%d}. Please select another date",
            )
            return val_date

    @cached_property
    def base_df(self) -> pd.DataFrame:
        if self.action != "metadata":
            with suppress(KeyError):
                df = (
                    pd.DataFrame(
                        AssetPosition.objects.filter(portfolio=self.portfolio, date=self.val_date)
                        .exclude(Q(underlying_quote__is_cash=True) | Q(underlying_quote__is_cash_equivalent=True))
                        .values("underlying_quote", "weighting", "total_value_fx_usd")
                    )
                    .groupby("underlying_quote")
                    .sum()
                    .astype(float)
                )
                df["weighting"] = df["weighting"] / df["weighting"].sum()
                return df
        return pd.DataFrame(columns=["weighting", "total_value_fx_usd"])

    @cached_property
    def esg_aggregation(self) -> ESGAggregation | None:
        return ESGAggregation[self.request.GET.get("esg_aggregation", ESGAggregation.GHG_EMISSIONS_SCOPE_1.name)]

    @cached_property
    def esg_data(self) -> pd.Series:
        if self.action != "metadata":
            return self.esg_aggregation.get_esg_data(Instrument.objects.filter(id__in=self.base_df.index))
        return pd.Series(dtype="float64")

    @cached_property
    def dataloader(self) -> DataLoader:
        dataloader = DataLoader(
            self.base_df["weighting"],
            self.esg_data,
            self.val_date,
            total_value_fx_usd=self.base_df["total_value_fx_usd"],
        )
        self.metric = dataloader.compute(self.esg_aggregation).rename("metric")
        return dataloader

    @cached_property
    def metric_columns(self) -> list[str]:
        return (
            ["weighting", "total_value_fx_usd", "esg_data", "weights_in_coverage", "metric"]
            + list(map(lambda x: x.series.name, self.dataloader.extra_esg_data_logs))
            + list(map(lambda x: x.series.name, self.dataloader.intermediary_logs))
        )

    def get_pandas_fields(self, request):
        fields = [
            pf.PKField(key="underlying_instrument", label="ID"),
            pf.CharField(key="underlying_instrument_repr", label="Instrument"),
            pf.FloatField(key="weighting", label="Weight", percent=True),
            pf.FloatField(key="total_value_fx_usd", label="Total Asset (USD)"),
            pf.FloatField(key="esg_data", label="ESG Data", precision=4),
            pf.FloatField(key="weights_in_coverage", label="Weights in Coverage", percent=True),
        ]
        for extra_esg_data_log in self.dataloader.extra_esg_data_logs:
            fields.append(
                pf.FloatField(
                    key=extra_esg_data_log.series.name,
                    label=extra_esg_data_log.label,
                    percent=extra_esg_data_log.is_percent,
                    precision=extra_esg_data_log.precision,
                )
            )
        for intermediary_log in self.dataloader.intermediary_logs:
            fields.append(
                pf.FloatField(
                    key=intermediary_log.series.name,
                    label=intermediary_log.label,
                    percent=intermediary_log.is_percent,
                    precision=intermediary_log.precision,
                )
            )
        if self.esg_aggregation.get_aggregation() == AggregationMethod.PERCENTAGE_SUM:
            fields.append(pf.FloatField(key="metric", label="Metric", percent=True))
        else:
            fields.append(pf.FloatField(key="metric", label="Metric", precision=4))
        return pf.PandasFields(fields=fields)

    def get_dataframe(self, request, queryset, **kwargs):
        dataloader = self.dataloader
        df = pd.concat(
            [
                self.base_df["weighting"],
                self.base_df["total_value_fx_usd"],
                self.esg_data.rename("esg_data"),
                *map(lambda x: x.series, dataloader.extra_esg_data_logs),
                dataloader.weights_in_coverage.rename("weights_in_coverage"),
                *map(lambda x: x.series, dataloader.intermediary_logs),
                self.metric,
            ],
            axis=1,
        )
        return df

    def manipulate_dataframe(self, df):
        df["underlying_instrument_repr"] = df.index.map(
            dict(self.dataloader.instruments.values_list("id", "computed_str"))
        )
        return df.reset_index()

    def get_aggregates(self, request, df):
        aggregates = {}
        for column in self.metric_columns:
            with suppress(TypeError, KeyError):  # silent aggregation on str type
                aggregates[column] = {"Σ": format_number(df[column].sum(), decimal=4)}
        return aggregates

    def get_ordering_fields(self) -> list[str]:
        return ["underlying_instrument_repr"] + self.metric_columns
